"""Test endpoint for envoy v7 and newer firmware"""

import contextlib
import json
import logging
from typing import Any

import aiohttp
import orjson
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion

from pyenphase.const import (
    PhaseNames,
)
from pyenphase.envoy import EnvoyProbeFailed, SupportedFeatures

from .common import (
    endpoint_path,
    get_mock_envoy,
    latest_request,
    load_json_fixture,
    override_mock,
    prep_envoy,
    start_7_firmware_mock,
    temporary_log_level,
    updater_features,
)

# we're testing, ignore private member access
# pyright: reportPrivateUsage=false

LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize(
    (
        "version",
        "part_number",
        "supported_features",
        "updaters",
        "phase_count",
        "production_phases",
        "consumption_phases",
    ),
    [
        (
            "5.0.62",
            "800-00551-r02",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "4.10.35",
            "800-00555-r03",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
            },
            2,
            {},
            {},
        ),
        (
            "7.3.130",
            "800-00555-r03",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
            },
            1,
            {},
            {},
        ),
        (
            "7.3.130_no_consumption",
            "800-00647-r10",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
            },
            2,
            {},
            {},
        ),
        (
            "7.3.517",
            "800-00555-r03",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyEnembleUpdater": SupportedFeatures.ENPOWER
                | SupportedFeatures.ENCHARGE,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "7.3.517_legacy_savings_mode",
            "800-00555-r03",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyEnembleUpdater": SupportedFeatures.ENPOWER
                | SupportedFeatures.ENCHARGE,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "7.3.517_system_2",
            "800-00555-r03",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyEnembleUpdater": SupportedFeatures.ENPOWER
                | SupportedFeatures.ENCHARGE,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
            },
            2,
            {},
            {},
        ),
        (
            "7.3.466_metered_disabled_cts",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.114_without_cts",
            "800-00656-r06",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.175",
            "800-00555-r03",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.175_total",
            "800-00654-r06",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.175_standard",
            "800-00656-r06",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.175_with_cts",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
            },
            1,
            {},
            {},
        ),
        (
            "7.6.175_with_cts_3phase",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.THREEPHASE
                | SupportedFeatures.CTMETERS,
            },
            3,
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 1869678,
                    "watt_hours_last_7_days": 29891,
                    "watt_hours_today": 2200,
                    "watts_now": -3,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 1241246,
                    "watt_hours_last_7_days": 19794,
                    "watt_hours_today": 1455,
                    "watts_now": 0,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 1240189,
                    "watt_hours_last_7_days": 19807,
                    "watt_hours_today": 1458,
                    "watts_now": -4,
                },
            },
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 2293783,
                    "watt_hours_last_7_days": 39392,
                    "watt_hours_today": 8585,
                    "watts_now": 89,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 948058,
                    "watt_hours_last_7_days": 18949,
                    "watt_hours_today": 2155,
                    "watts_now": 123,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 832954,
                    "watt_hours_last_7_days": 10443,
                    "watt_hours_today": 1683,
                    "watts_now": -3,
                },
            },
        ),
        (
            "7.3.466_with_cts_3phase",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.THREEPHASE
                | SupportedFeatures.CTMETERS,
            },
            3,
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 1869678,
                    "watt_hours_last_7_days": 29891,
                    "watt_hours_today": 2200,
                    "watts_now": -3,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 1241246,
                    "watt_hours_last_7_days": 19794,
                    "watt_hours_today": 1455,
                    "watts_now": 0,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 1240189,
                    "watt_hours_last_7_days": 19807,
                    "watt_hours_today": 1458,
                    "watts_now": -4,
                },
            },
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 2293783,
                    "watt_hours_last_7_days": 39392,
                    "watt_hours_today": 8585,
                    "watts_now": 89,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 948058,
                    "watt_hours_last_7_days": 18949,
                    "watt_hours_today": 2155,
                    "watts_now": 123,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 832954,
                    "watt_hours_last_7_days": 10443,
                    "watt_hours_today": 1683,
                    "watts_now": -3,
                },
            },
        ),
        (
            "7.6.185_with_cts_and_battery_3t",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
            },
            1,
            {},
            {},
        ),
        (
            "8.1.41",
            "800-00664-r05",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.PRODUCTION,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENPOWER
                | SupportedFeatures.ENCHARGE,
                "EnvoyProductionJsonUpdater": SupportedFeatures.PRODUCTION
                | SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION,
            },
            1,
            {},
            {},
        ),
        (
            "8.2.127_with_3cts_and_battery_split",
            "800-00654-r08",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE
                | SupportedFeatures.ENPOWER,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
            },
            2,
            {},
            {},
        ),
        (
            "8.2.127_with_generator_running",
            "800-00647-r09",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS
            | SupportedFeatures.GENERATOR,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE
                | SupportedFeatures.ENPOWER,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
                "EnvoyGeneratorUpdater": SupportedFeatures.GENERATOR,
            },
            2,
            {},
            {},
        ),
        (
            "8.2.4286_with_3cts_and_battery_split",
            "800-00664-r05",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE
                | SupportedFeatures.ENPOWER,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
                "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE
                | SupportedFeatures.CTMETERS,
            },
            2,
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 6709433,
                    "watt_hours_last_7_days": 6703259,
                    "watt_hours_today": 6277,
                    "watts_now": 3559,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 6721896,
                    "watt_hours_last_7_days": 6715706,
                    "watt_hours_today": 6293,
                    "watts_now": 3564,
                },
            },
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 7197821,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 4407,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 7915653,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 4478,
                },
            },
        ),
        (
            "8.2.4264_metered_noct",
            "800-00554-r03",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            {
                "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
                "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
        (
            "8.2.4345_with_device_data",
            "800-00649-r01",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS
            | SupportedFeatures.DETAILED_INVERTERS,
            {
                "EnvoyDeviceDataInvertersUpdater": SupportedFeatures.INVERTERS
                | SupportedFeatures.DETAILED_INVERTERS,
                "EnvoyEnembleUpdater": SupportedFeatures.ENCHARGE
                | SupportedFeatures.ENPOWER,
                "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
                "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
                | SupportedFeatures.TOTAL_CONSUMPTION
                | SupportedFeatures.NET_CONSUMPTION
                | SupportedFeatures.PRODUCTION,
                "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
            },
            1,
            {},
            {},
        ),
    ],
    ids=[
        "5.0.62",
        "4.10.35",
        "7.3.130",
        "7.3.130_no_consumption",
        "7.3.517",
        "7.3.517_legacy_savings_mode",
        "7.3.517_system_2",
        "7.3.466_metered_disabled_cts",
        "7.6.114_without_cts",
        "7.6.175",
        "7.6.175_total",
        "7.6.175_standard",
        "7.6.175_with_cts",
        "7.6.175_with_cts_3phase",
        "7.3.466_with_cts_3phase",
        "7.6.185_with_cts_and_battery_3t",
        "8.1.41",
        "8.2.127_with_3cts_and_battery_split",
        "8.2.127_with_generator_running",
        "8.2.4286_with_3cts_and_battery_split",
        "8.2.4264_metered_noct",
        "8.2.4345_with_device_data",
    ],
)
@pytest.mark.asyncio
async def test_with_7_x_firmware(
    version: str,
    part_number: str,
    snapshot: SnapshotAssertion,
    supported_features: SupportedFeatures,
    updaters: dict[str, SupportedFeatures],
    caplog: pytest.LogCaptureFixture,
    phase_count: int,
    production_phases: dict[str, dict[str, Any]],
    consumption_phases: dict[str, dict[str, Any]],
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 7.x firmware."""
    start_7_firmware_mock(mock_aioresponse)
    files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    caplog.set_level(logging.DEBUG)

    envoy = await get_mock_envoy(test_client_session)
    # get http or https paths based on firmware version
    full_host = endpoint_path(version, envoy.host)
    data = envoy.data
    assert data is not None
    assert data == snapshot

    # verify expected properties
    assert envoy.firmware == version.split("_")[0]
    assert envoy.serial_number

    assert envoy.part_number == part_number
    assert updater_features(envoy._updaters) == updaters
    assert envoy._supported_features == supported_features
    assert envoy.phase_count == phase_count

    # test envoy request methods GET, PUT and POST
    test_data = await load_json_fixture(version, "api_v1_production_inverters")
    mock_aioresponse.post(
        f"{full_host}/api/v1/production/inverters",
        status=200,
        payload=test_data,
        repeat=True,
    )
    mock_aioresponse.put(
        f"{full_host}/api/v1/production/inverters",
        status=200,
        payload=test_data,
        repeat=True,
    )

    # set log level to info 1 time for GET and 1 time for POST to improve COV
    with temporary_log_level("pyenphase", logging.INFO):
        # test request with just an endpoint, should be a GET
        myresponse: aiohttp.ClientResponse = await envoy.request(
            "/api/v1/production/inverters"
        )
        # Check that at least one GET request was made to this URL
        cnt, request_data = latest_request(
            mock_aioresponse, "GET", "/api/v1/production/inverters"
        )
        assert cnt > 0
        assert await myresponse.json() == test_data

        # with data but no method should be post
        await envoy.request("/api/v1/production/inverters", data=test_data)
        # Check that at least one POST request was made to this URL
        cnt, request_data = latest_request(
            mock_aioresponse, "POST", "/api/v1/production/inverters"
        )
        assert cnt == 1
        assert orjson.loads(request_data) == test_data

    # with method should be specified method
    caplog.clear()
    await envoy.request("/api/v1/production/inverters", data=test_data, method="PUT")
    # Check that at least one PUT request was made to this URL
    cnt, request_data = latest_request(
        mock_aioresponse, "PUT", "/api/v1/production/inverters"
    )
    assert cnt == 1
    assert orjson.loads(request_data) == test_data
    # verify debug log shows correct method
    assert "Sending PUT to" in caplog.text

    # change data to recognize this from previous requests
    caplog.clear()
    test_data.append({"second_post": "test"})  # type: ignore[attr-defined]
    await envoy.request("/api/v1/production/inverters", data=test_data, method="POST")
    # Check that POST requests with changed data was made
    cnt, request_data = latest_request(
        mock_aioresponse, "POST", "/api/v1/production/inverters"
    )
    assert cnt == 1
    assert orjson.loads(request_data) == test_data
    # verify debug log shows correct method
    assert "Sending POST to" in caplog.text
    caplog.clear()

    # data is the original collected envoy.data
    # are all production phases reported
    expected_phases = production_phases == {}
    actual_phases = data.system_production_phases is None
    assert not (expected_phases ^ actual_phases)

    # are all consumption phases reported
    expected_phases = consumption_phases == {}
    actual_phases = data.system_consumption_phases is None
    assert not (expected_phases ^ actual_phases)

    reported_phase_count = envoy.active_phase_count
    # are all production phases reported
    expected_phase_count = len(production_phases)
    assert expected_phase_count == reported_phase_count

    # are all consumption phases reported
    expected_phase_count = len(consumption_phases)
    assert expected_phase_count == reported_phase_count

    # Test each production phase
    for phase in production_phases:
        assert data.system_production_phases is not None
        proddata = data.system_production_phases[phase]
        assert proddata is not None
        modeldata = production_phases[phase]

        # test each element of the phase data
        assert proddata.watt_hours_lifetime == modeldata["watt_hours_lifetime"]
        assert proddata.watt_hours_last_7_days == modeldata["watt_hours_last_7_days"]
        assert proddata.watt_hours_today == modeldata["watt_hours_today"]
        assert proddata.watts_now == modeldata["watts_now"]

    # are all consumption phases reported
    assert (
        envoy.active_phase_count == 0
        if data.system_consumption_phases is None
        else len(data.system_consumption_phases)
    )
    # Test each consumption phase
    for phase in consumption_phases:
        assert data.system_consumption_phases is not None
        consdata = data.system_consumption_phases[phase]
        assert consdata is not None
        modeldata = consumption_phases[phase]

        # test each element of the phase data
        assert consdata.watt_hours_lifetime == modeldata["watt_hours_lifetime"]
        assert consdata.watt_hours_last_7_days == modeldata["watt_hours_last_7_days"]
        assert consdata.watt_hours_today == modeldata["watt_hours_today"]
        assert consdata.watts_now == modeldata["watts_now"]

    # COV test with no production segment
    if "production" in files:
        try:
            json_data = await load_json_fixture(version, "production")
        except json.decoder.JSONDecodeError:
            json_data = None
        if json_data:
            del json_data["production"]
        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/production",
            status=200,
            payload=json_data,
        )
    else:
        override_mock(mock_aioresponse, "get", f"{full_host}/production", status=404)
    with contextlib.suppress(EnvoyProbeFailed):
        await envoy.probe()

    # test inverter device data with missing data fields
    if "ivp_pdm_device_data" in files:
        # rebuild default data setup
        files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
        await envoy.probe()

        # verify we have inverter and inverter details details
        assert envoy.supported_features & (
            SupportedFeatures.DETAILED_INVERTERS | SupportedFeatures.INVERTERS
        )
        json_data = await load_json_fixture(version, "ivp_pdm_device_data")

        # remove channels from first inverter, should cause switch to production inverter data
        for key in json_data:
            if key not in ("deviceCount", "deviceDataLimit"):
                del json_data[key]["channels"]
                break

        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/ivp/pdm/device_data",
            status=200,
            payload=json_data,
        )
        await envoy.probe()
        # verify we have production inverter data only
        assert envoy.supported_features & SupportedFeatures.INVERTERS
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].ac_frequency is None

        # rebuild default data setup
        files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
        await envoy.probe()

        # verify we have inverter and inverter details details
        assert envoy.supported_features & (
            SupportedFeatures.DETAILED_INVERTERS | SupportedFeatures.INVERTERS
        )
        json_data = await load_json_fixture(version, "ivp_pdm_device_data")

        # remove lastReading from first inverter channel, should cause switch to production inverter data
        for key in json_data:
            if key not in ("deviceCount", "deviceDataLimit"):
                del json_data[key]["channels"][0]["lastReading"]
                break

        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/ivp/pdm/device_data",
            status=200,
            payload=json_data,
        )
        await envoy.probe()
        # verify we have production inverter data only
        assert envoy.supported_features & SupportedFeatures.INVERTERS
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].ac_frequency is None

        # rebuild default data setup
        files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
        await envoy.probe()

        # verify we have inverter and inverter details
        assert envoy.supported_features & (
            SupportedFeatures.DETAILED_INVERTERS | SupportedFeatures.INVERTERS
        )
        json_data = await load_json_fixture(version, "ivp_pdm_device_data")

        # set deviceCount equal to deviceDataLimit, should cause switch to production inverter data
        json_data["deviceCount"] = json_data["deviceDataLimit"]

        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/ivp/pdm/device_data",
            status=200,
            payload=json_data,
        )
        await envoy.probe()
        # verify we have production inverter data only
        assert envoy.supported_features & SupportedFeatures.INVERTERS
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].ac_frequency is None

        # rebuild default data setup
        files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
        await envoy.probe()

        # verify we have inverter and inverter details
        assert envoy.supported_features & (
            SupportedFeatures.DETAILED_INVERTERS | SupportedFeatures.INVERTERS
        )
        json_data = await load_json_fixture(version, "ivp_pdm_device_data")

        # set deviceCount equal to deviceDataLimit, should cause switch to production inverter data
        json_data["deviceCount"] = json_data["deviceDataLimit"]

        # remove deviceCount key, should cause switch to production inverter data
        del json_data["deviceCount"]

        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/ivp/pdm/device_data",
            status=200,
            payload=json_data,
        )
        await envoy.probe()
        # verify we have production inverter data only
        assert envoy.supported_features & SupportedFeatures.INVERTERS
        assert not (envoy.supported_features & SupportedFeatures.DETAILED_INVERTERS)
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].ac_frequency is None

        # rebuild default data setup
        files = await prep_envoy(mock_aioresponse, "127.0.0.1", version)
        await envoy.probe()

        # verify we have inverter and inverter details details
        assert envoy.supported_features & (
            SupportedFeatures.DETAILED_INVERTERS | SupportedFeatures.INVERTERS
        )
        json_data = await load_json_fixture(version, "ivp_pdm_device_data")

        # remove lifetime from first channel of inverters, should cause lifetime to be None
        for key in json_data:
            if key not in ("deviceCount", "deviceDataLimit"):
                del json_data[key]["channels"][0]["lifetime"]

        override_mock(
            mock_aioresponse,
            "get",
            f"{full_host}/ivp/pdm/device_data",
            status=200,
            payload=json_data,
            repeat=2,
        )
        await envoy.probe()
        # verify we have still have inverter details
        assert (
            envoy.supported_features & SupportedFeatures.INVERTERS
            | SupportedFeatures.INVERTERS
        )
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].lifetime_energy is None
            assert data.inverters[key].last_report_watts is not None

    else:
        override_mock(
            mock_aioresponse, "get", f"{full_host}/ivp/pdm/device_data", status=404
        )
        await envoy.probe()
        data = await envoy.update()
        for key in data.inverters:
            assert data.inverters[key].ac_frequency is None



"""Test envoy firmware prior to v7."""

import logging
import re

import aiohttp
import pytest
from aioresponses import aioresponses

from pyenphase import EnvoyInverter, register_updater
from pyenphase.const import URL_PRODUCTION
from pyenphase.envoy import SupportedFeatures, get_updaters
from pyenphase.exceptions import (
    ENDPOINT_PROBE_EXCEPTIONS,
    EnvoyAuthenticationRequired,
    EnvoyFeatureNotAvailable,
    EnvoyPoorDataQuality,
    EnvoyProbeFailed,
)
from pyenphase.models.envoy import EnvoyData
from pyenphase.models.meters import EnvoyPhaseMode
from pyenphase.models.system_production import EnvoySystemProduction
from pyenphase.updaters.base import EnvoyUpdater

from .common import (
    get_mock_envoy,
    load_json_fixture,
    override_mock,
    prep_envoy,
    updater_features,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_with_4_2_27_firmware(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 4.2.27 firmware."""
    version = "4.2.27"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data: EnvoyData | None = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.METERING)
    assert not (envoy._supported_features & SupportedFeatures.INVERTERS)
    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00551-r02"

    assert data.system_production is not None
    assert (
        data.system_production.watts_now == 5894
    )  # This used to use the production.json endpoint, but its always a bit behind
    assert data.system_production.watt_hours_today == 17920
    assert data.system_production.watt_hours_last_7_days == 276614
    assert data.system_production.watt_hours_lifetime == 10279087
    assert not data.inverters
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert envoy.envoy_model == "Envoy"

    # Test that Ensemble commands raise FeatureNotAvailable
    with pytest.raises(EnvoyFeatureNotAvailable):
        await envoy.go_off_grid()
    with pytest.raises(EnvoyFeatureNotAvailable):
        await envoy.go_on_grid()


@pytest.mark.asyncio
async def test_with_4_2_33_firmware_no_cons_ct(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 4.2.33 firmware."""
    version = "4.2.33"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data: EnvoyData | None = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert envoy._supported_features & SupportedFeatures.METERING
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert updater_features(envoy._updaters) == {
        "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
        | SupportedFeatures.PRODUCTION,
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyMetersUpdater": SupportedFeatures.DUALPHASE | SupportedFeatures.CTMETERS,
    }
    assert envoy.part_number == "800-00547-r05"

    assert data.system_production is not None
    assert (
        data.system_production.watts_now == -9
    )  # This used to use the production.json endpoint, but its always a bit behind
    assert data.system_production.watt_hours_today == 10215
    assert data.system_production.watt_hours_last_7_days == 10833
    assert data.system_production.watt_hours_lifetime == 8598257
    assert not data.system_consumption
    assert data.inverters["1234567890"] == EnvoyInverter(
        serial_number="1234567890",
        last_report_date=1743551631,
        last_report_watts=3,
        max_report_watts=131,
    )
    assert data.inverters["121622033019"] == EnvoyInverter(
        serial_number="121622033019",
        last_report_date=1536668634,
        last_report_watts=17,
        max_report_watts=17,
    )
    assert envoy.ct_meter_count == 1
    assert envoy.phase_count == 2
    assert envoy.phase_mode == EnvoyPhaseMode.SPLIT
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert envoy.envoy_model == "Envoy, phases: 2, phase mode: split, production CT"

    # Test that Ensemble commands raise FeatureNotAvailable
    with pytest.raises(EnvoyFeatureNotAvailable):
        await envoy.go_off_grid()
    with pytest.raises(EnvoyFeatureNotAvailable):
        await envoy.go_on_grid()

    await envoy.update()


@pytest.mark.asyncio
async def test_with_5_0_49_firmware(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 5.0.49 firmware."""
    version = "5.0.49"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00551-r02"
    assert envoy.phase_count == 1

    assert not data.system_consumption
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert data.system_production is not None
    assert data.system_production.watts_now == 4859
    assert data.system_production.watt_hours_today == 5046
    assert data.system_production.watt_hours_last_7_days == 445686
    assert data.system_production.watt_hours_lifetime == 88742152
    assert data.inverters == {
        "121547055830": EnvoyInverter(
            serial_number="121547055830",
            last_report_date=1618083280,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059008": EnvoyInverter(
            serial_number="121547059008",
            last_report_date=1618083240,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547059079": EnvoyInverter(
            serial_number="121547059079",
            last_report_date=1618083244,
            last_report_watts=130,
            max_report_watts=257,
        ),
        "121547059102": EnvoyInverter(
            serial_number="121547059102",
            last_report_date=1618083273,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547059107": EnvoyInverter(
            serial_number="121547059107",
            last_report_date=1618083265,
            last_report_watts=132,
            max_report_watts=257,
        ),
        "121547059108": EnvoyInverter(
            serial_number="121547059108",
            last_report_date=1618083266,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059112": EnvoyInverter(
            serial_number="121547059112",
            last_report_date=1618083286,
            last_report_watts=132,
            max_report_watts=257,
        ),
        "121547059128": EnvoyInverter(
            serial_number="121547059128",
            last_report_date=1618083262,
            last_report_watts=135,
            max_report_watts=257,
        ),
        "121547059193": EnvoyInverter(
            serial_number="121547059193",
            last_report_date=1618083250,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059202": EnvoyInverter(
            serial_number="121547059202",
            last_report_date=1618083251,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547059217": EnvoyInverter(
            serial_number="121547059217",
            last_report_date=1618083281,
            last_report_watts=137,
            max_report_watts=257,
        ),
        "121547059253": EnvoyInverter(
            serial_number="121547059253",
            last_report_date=1618083289,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059333": EnvoyInverter(
            serial_number="121547059333",
            last_report_date=1618083277,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547059354": EnvoyInverter(
            serial_number="121547059354",
            last_report_date=1618083287,
            last_report_watts=132,
            max_report_watts=257,
        ),
        "121547059355": EnvoyInverter(
            serial_number="121547059355",
            last_report_date=1618083263,
            last_report_watts=131,
            max_report_watts=258,
        ),
        "121547059357": EnvoyInverter(
            serial_number="121547059357",
            last_report_date=1618083254,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547059359": EnvoyInverter(
            serial_number="121547059359",
            last_report_date=1618083247,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547059360": EnvoyInverter(
            serial_number="121547059360",
            last_report_date=1618083245,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059363": EnvoyInverter(
            serial_number="121547059363",
            last_report_date=1618083255,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547059381": EnvoyInverter(
            serial_number="121547059381",
            last_report_date=1618083259,
            last_report_watts=132,
            max_report_watts=257,
        ),
        "121547059889": EnvoyInverter(
            serial_number="121547059889",
            last_report_date=1618083264,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060383": EnvoyInverter(
            serial_number="121547060383",
            last_report_date=1618083257,
            last_report_watts=135,
            max_report_watts=258,
        ),
        "121547060384": EnvoyInverter(
            serial_number="121547060384",
            last_report_date=1618083250,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547060392": EnvoyInverter(
            serial_number="121547060392",
            last_report_date=1618083288,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060396": EnvoyInverter(
            serial_number="121547060396",
            last_report_date=1618083269,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060412": EnvoyInverter(
            serial_number="121547060412",
            last_report_date=1618083258,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547060415": EnvoyInverter(
            serial_number="121547060415",
            last_report_date=1618083267,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060590": EnvoyInverter(
            serial_number="121547060590",
            last_report_date=1618083277,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060592": EnvoyInverter(
            serial_number="121547060592",
            last_report_date=1618083279,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060593": EnvoyInverter(
            serial_number="121547060593",
            last_report_date=1618083271,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060643": EnvoyInverter(
            serial_number="121547060643",
            last_report_date=1618083284,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547060647": EnvoyInverter(
            serial_number="121547060647",
            last_report_date=1618083285,
            last_report_watts=134,
            max_report_watts=258,
        ),
        "121547060650": EnvoyInverter(
            serial_number="121547060650",
            last_report_date=1618083253,
            last_report_watts=131,
            max_report_watts=257,
        ),
        "121547060670": EnvoyInverter(
            serial_number="121547060670",
            last_report_date=1618083270,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547060671": EnvoyInverter(
            serial_number="121547060671",
            last_report_date=1618083283,
            last_report_watts=135,
            max_report_watts=257,
        ),
        "121547060727": EnvoyInverter(
            serial_number="121547060727",
            last_report_date=1618083275,
            last_report_watts=134,
            max_report_watts=257,
        ),
        "121547060758": EnvoyInverter(
            serial_number="121547060758",
            last_report_date=1618083274,
            last_report_watts=130,
            max_report_watts=255,
        ),
        "121547060761": EnvoyInverter(
            serial_number="121547060761",
            last_report_date=1618083260,
            last_report_watts=133,
            max_report_watts=257,
        ),
        "121547060766": EnvoyInverter(
            serial_number="121547060766",
            last_report_date=1618083242,
            last_report_watts=132,
            max_report_watts=257,
        ),
        "121547060773": EnvoyInverter(
            serial_number="121547060773",
            last_report_date=1618083247,
            last_report_watts=132,
            max_report_watts=257,
        ),
    }
    assert envoy.envoy_model == "Envoy"


@pytest.mark.asyncio
async def test_with_3_7_0_firmware(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.7.0 firmware."""
    version = "3.7.0"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # Verify the library does not support scraping to comply with ADR004
    with pytest.raises(EnvoyProbeFailed):
        await get_mock_envoy(test_client_session)

    # Test the register interface by registering a legacy production scraper
    #
    # ADR004 compliance:
    # We won't do this in our production code as we don't support scraping, but
    # we want to leave the door open for custom components to use the interface.
    #

    _KEY_TO_REGEX = {
        "watts_now": r"<td>Current.*</td>\s*<td>\s*(\d+|\d+\.\d+)\s*(W|kW|MW)</td>",
        "watt_hours_last_7_days": r"<td>Past Week</td>\s*<td>\s*(\d+|\d+\.\d+)\s*(Wh|kWh|MWh)</td>",
        "watt_hours_today": r"<td>Today</td>\s*<td>\s*(\d+|\d+\.\d+)\s*(Wh|kWh|MWh)</td>",
        "watt_hours_lifetime": r"<td>Since Installation</td>\s*<td>\s*(\d+|\d+\.\d+)\s*(Wh|kWh|MWh)</td>",
    }

    class LegacyEnvoySystemProduction(EnvoySystemProduction):
        @classmethod
        def from_production_legacy(cls, text: str) -> EnvoySystemProduction:
            """Legacy parser."""
            data: dict[str, int] = {
                "watts_now": 0,
                "watt_hours_today": 0,
                "watt_hours_last_7_days": 0,
                "watt_hours_lifetime": 0,
            }
            for key, regex in _KEY_TO_REGEX.items():
                if match := re.search(regex, text, re.MULTILINE):
                    unit = match.group(2).lower()
                    value = float(match.group(1))
                    if unit.startswith("k"):
                        value *= 1000
                    elif unit.startswith("m"):
                        value *= 1000000
                    data[key] = int(value)
            return cls(**data)

    class LegacyProductionScraper(EnvoyUpdater):
        async def probe(
            self, discovered_features: SupportedFeatures
        ) -> SupportedFeatures | None:
            """Probe the Envoy for this updater and return SupportedFeatures."""
            if SupportedFeatures.PRODUCTION in discovered_features:
                # Already discovered from another updater
                return None

            try:
                response = await self._probe_request(URL_PRODUCTION)
                data = await response.text()
            except ENDPOINT_PROBE_EXCEPTIONS:
                return None
            if "Since Installation" not in data:
                return None
            self._supported_features |= SupportedFeatures.PRODUCTION
            return self._supported_features

        async def update(self, envoy_data: EnvoyData) -> None:
            """Update the Envoy for this updater."""
            response = await self._request(URL_PRODUCTION)
            production_data = await response.text()
            envoy_data.raw[URL_PRODUCTION] = production_data
            envoy_data.system_production = (
                LegacyEnvoySystemProduction.from_production_legacy(production_data)
            )

    remove = register_updater(LegacyProductionScraper)
    assert LegacyProductionScraper in get_updaters()
    try:
        envoy = await get_mock_envoy(test_client_session)
        data = envoy.data
        assert data is not None
        assert envoy._supported_features is not None

        assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
        assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
        assert not (envoy._supported_features & SupportedFeatures.INVERTERS)
        assert updater_features(envoy._updaters) == {
            "LegacyProductionScraper": SupportedFeatures.PRODUCTION,
        }
        assert envoy.part_number == "800-00069-r05"

        assert not data.system_consumption
        assert data.system_production is not None
        assert data.system_production.watts_now == 6630
        assert data.system_production.watt_hours_today == 53600
        assert data.system_production.watt_hours_last_7_days == 405000
        assert data.system_production.watt_hours_lifetime == 133000000
        assert not data.inverters
        assert envoy.ct_meter_count == 0
        assert envoy.phase_count == 1
        assert envoy.phase_mode is None
        assert envoy.consumption_meter_type is None
        assert not data.system_consumption_phases
        assert not data.system_production_phases
        assert envoy.envoy_model == "Envoy"
    finally:
        remove()
        assert LegacyProductionScraper not in get_updaters()


@pytest.mark.asyncio
async def test_with_3_9_36_firmware_bad_auth(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.9.36 firmware with incorrect auth."""
    version = "3.9.36_bad_auth"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # Override production endpoints to return 401
    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/api/v1/production",
        status=401,
        payload={
            "status": 401,
            "error": "",
            "info": "Authentication required",
            "moreInfo": "",
        },
        repeat=True,
    )

    with pytest.raises(EnvoyAuthenticationRequired):
        await get_mock_envoy(test_client_session)


@pytest.mark.asyncio
async def test_with_3_9_36_firmware_no_inverters(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.9.36 firmware with auth that does not allow inverters."""
    version = "3.9.36_bad_auth"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    # force auth failure on inverters
    mock_aioresponse.get(
        "http://127.0.0.1/api/v1/production/inverters",
        status=401,
        payload=await load_json_fixture(version, "api_v1_production_inverters"),
        repeat=True,
    )

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.INVERTERS)
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00069-r05"
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases


@pytest.mark.asyncio
async def test_with_3_9_36_firmware(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.9.36 firmware."""
    version = "3.9.36"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    # no access to tariff
    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/admin/lib/tariff",
        status=401,
        repeat=True,
    )

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00069-r05"

    assert not data.system_consumption
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert data.system_production is not None
    assert data.system_production.watts_now == 1271
    assert data.system_production.watt_hours_today == 1460
    assert data.system_production.watt_hours_last_7_days == 130349
    assert data.system_production.watt_hours_lifetime == 6012540
    assert data.inverters == {
        "121547058983": EnvoyInverter(
            serial_number="121547058983",
            last_report_date=1618083969,
            last_report_watts=137,
            max_report_watts=238,
        ),
        "121547058993": EnvoyInverter(
            serial_number="121547058993",
            last_report_date=1618083961,
            last_report_watts=138,
            max_report_watts=231,
        ),
        "121547060394": EnvoyInverter(
            serial_number="121547060394",
            last_report_date=1618083966,
            last_report_watts=138,
            max_report_watts=238,
        ),
        "121547060402": EnvoyInverter(
            serial_number="121547060402",
            last_report_date=1618083962,
            last_report_watts=138,
            max_report_watts=240,
        ),
        "121547060495": EnvoyInverter(
            serial_number="121547060495",
            last_report_date=1618083959,
            last_report_watts=135,
            max_report_watts=228,
        ),
        "121547060638": EnvoyInverter(
            serial_number="121547060638",
            last_report_date=1618083966,
            last_report_watts=139,
            max_report_watts=241,
        ),
        "121547060646": EnvoyInverter(
            serial_number="121547060646",
            last_report_date=1618083957,
            last_report_watts=139,
            max_report_watts=240,
        ),
        "121547060652": EnvoyInverter(
            serial_number="121547060652",
            last_report_date=1618083959,
            last_report_watts=140,
            max_report_watts=245,
        ),
        "121603025842": EnvoyInverter(
            serial_number="121603025842",
            last_report_date=1618083963,
            last_report_watts=139,
            max_report_watts=260,
        ),
        "121603034267": EnvoyInverter(
            serial_number="121603034267",
            last_report_date=1618083956,
            last_report_watts=138,
            max_report_watts=244,
        ),
        "121603038867": EnvoyInverter(
            serial_number="121603038867",
            last_report_date=1618083964,
            last_report_watts=138,
            max_report_watts=242,
        ),
        "121603039216": EnvoyInverter(
            serial_number="121603039216",
            last_report_date=1618083968,
            last_report_watts=139,
            max_report_watts=273,
        ),
    }
    assert envoy.envoy_model == "Envoy"


@pytest.mark.asyncio
async def test_with_3_9_36_firmware_with_production_401(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.9.36 firmware when /production throws a 401."""
    version = "3.9.36"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    # force 401 on production
    mock_aioresponse.get("http://127.0.0.1/production", status=401, repeat=True)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00069-r05"

    assert not data.system_consumption
    assert data.system_production is not None
    assert data.system_production.watts_now == 1271
    assert data.system_production.watt_hours_today == 1460
    assert data.system_production.watt_hours_last_7_days == 130349
    assert data.system_production.watt_hours_lifetime == 6012540
    assert data.inverters
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases


@pytest.mark.asyncio
async def test_with_3_9_36_firmware_with_production_and_production_json_401(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.9.36 firmware when /production and /production.json throws a 401."""
    version = "3.9.36"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    # force 401 on production
    override_mock(
        mock_aioresponse, "get", "http://127.0.0.1/production", status=401, repeat=True
    )
    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/production.json",
        status=401,
        repeat=True,
    )
    # Also need to override the API v1 endpoint
    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/api/v1/production",
        status=401,
        repeat=True,
    )

    with pytest.raises(EnvoyAuthenticationRequired):
        await get_mock_envoy(test_client_session)


@pytest.mark.asyncio
async def test_with_3_8_10_firmware_with_meters_401(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Verify with 3.8.10 firmware when /ivp/meters throws a 401."""
    version = "3.8.10"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    override_mock(
        mock_aioresponse, "get", "http://127.0.0.1/ivp/meters", status=401, repeat=True
    )
    caplog.set_level(logging.DEBUG)
    await get_mock_envoy(test_client_session)
    assert "Skipping meters endpoint as user does not have access to" in caplog.text


@pytest.mark.asyncio
async def test_with_3_17_3_firmware(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.17.3 firmware."""
    version = "3.17.3"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }
    assert envoy.part_number == "800-00069-r05"

    assert not data.system_consumption
    assert envoy.ct_meter_count == 0
    assert envoy.phase_count == 1
    assert envoy.phase_mode is None
    assert envoy.consumption_meter_type is None
    assert not data.system_consumption_phases
    assert not data.system_production_phases
    assert data.system_production is not None
    assert data.system_production.watts_now == 5463
    assert data.system_production.watt_hours_today == 5481
    assert data.system_production.watt_hours_last_7_days == 389581
    assert data.system_production.watt_hours_lifetime == 93706280
    assert data.inverters == {
        "121512006273": EnvoyInverter(
            serial_number="121512006273",
            last_report_date=1618082959,
            last_report_watts=206,
            max_report_watts=254,
        ),
        "121512009183": EnvoyInverter(
            serial_number="121512009183",
            last_report_date=1618082961,
            last_report_watts=204,
            max_report_watts=253,
        ),
        "121512033008": EnvoyInverter(
            serial_number="121512033008",
            last_report_date=1618082947,
            last_report_watts=101,
            max_report_watts=243,
        ),
        "121512036220": EnvoyInverter(
            serial_number="121512036220",
            last_report_date=1618082927,
            last_report_watts=198,
            max_report_watts=245,
        ),
        "121512036221": EnvoyInverter(
            serial_number="121512036221",
            last_report_date=1618082963,
            last_report_watts=8,
            max_report_watts=116,
        ),
        "121512036250": EnvoyInverter(
            serial_number="121512036250",
            last_report_date=1618082940,
            last_report_watts=20,
            max_report_watts=190,
        ),
        "121512036336": EnvoyInverter(
            serial_number="121512036336",
            last_report_date=1618082932,
            last_report_watts=199,
            max_report_watts=247,
        ),
        "121512037453": EnvoyInverter(
            serial_number="121512037453",
            last_report_date=1618082949,
            last_report_watts=205,
            max_report_watts=255,
        ),
        "121512038416": EnvoyInverter(
            serial_number="121512038416",
            last_report_date=1618082953,
            last_report_watts=151,
            max_report_watts=251,
        ),
        "121512038421": EnvoyInverter(
            serial_number="121512038421",
            last_report_date=1618082949,
            last_report_watts=14,
            max_report_watts=233,
        ),
        "121512038619": EnvoyInverter(
            serial_number="121512038619",
            last_report_date=1618082962,
            last_report_watts=203,
            max_report_watts=252,
        ),
        "121512038691": EnvoyInverter(
            serial_number="121512038691",
            last_report_date=1618082942,
            last_report_watts=26,
            max_report_watts=247,
        ),
        "121512038762": EnvoyInverter(
            serial_number="121512038762",
            last_report_date=1618082930,
            last_report_watts=203,
            max_report_watts=253,
        ),
        "121512038845": EnvoyInverter(
            serial_number="121512038845",
            last_report_date=1618082945,
            last_report_watts=203,
            max_report_watts=253,
        ),
        "121512038901": EnvoyInverter(
            serial_number="121512038901",
            last_report_date=1618082944,
            last_report_watts=102,
            max_report_watts=245,
        ),
        "121512038919": EnvoyInverter(
            serial_number="121512038919",
            last_report_date=1618082959,
            last_report_watts=102,
            max_report_watts=238,
        ),
        "121512038982": EnvoyInverter(
            serial_number="121512038982",
            last_report_date=1618082950,
            last_report_watts=203,
            max_report_watts=253,
        ),
        "121512039005": EnvoyInverter(
            serial_number="121512039005",
            last_report_date=1618082933,
            last_report_watts=55,
            max_report_watts=254,
        ),
        "121512039018": EnvoyInverter(
            serial_number="121512039018",
            last_report_date=1618082964,
            last_report_watts=27,
            max_report_watts=252,
        ),
        "121512039075": EnvoyInverter(
            serial_number="121512039075",
            last_report_date=1618082930,
            last_report_watts=102,
            max_report_watts=237,
        ),
        "121512039090": EnvoyInverter(
            serial_number="121512039090",
            last_report_date=1618082946,
            last_report_watts=32,
            max_report_watts=194,
        ),
        "121512039091": EnvoyInverter(
            serial_number="121512039091",
            last_report_date=1618082939,
            last_report_watts=27,
            max_report_watts=252,
        ),
        "121512039093": EnvoyInverter(
            serial_number="121512039093",
            last_report_date=1618082966,
            last_report_watts=209,
            max_report_watts=256,
        ),
        "121512039124": EnvoyInverter(
            serial_number="121512039124",
            last_report_date=1618082938,
            last_report_watts=205,
            max_report_watts=254,
        ),
        "121512039143": EnvoyInverter(
            serial_number="121512039143",
            last_report_date=1618082956,
            last_report_watts=104,
            max_report_watts=245,
        ),
        "121512039181": EnvoyInverter(
            serial_number="121512039181",
            last_report_date=1618082943,
            last_report_watts=101,
            max_report_watts=238,
        ),
        "121512041456": EnvoyInverter(
            serial_number="121512041456",
            last_report_date=1618082937,
            last_report_watts=13,
            max_report_watts=79,
        ),
        "121512041640": EnvoyInverter(
            serial_number="121512041640",
            last_report_date=1618082927,
            last_report_watts=200,
            max_report_watts=249,
        ),
        "121512041747": EnvoyInverter(
            serial_number="121512041747",
            last_report_date=1618082925,
            last_report_watts=64,
            max_report_watts=248,
        ),
        "121512042132": EnvoyInverter(
            serial_number="121512042132",
            last_report_date=1618082924,
            last_report_watts=200,
            max_report_watts=250,
        ),
        "121512042344": EnvoyInverter(
            serial_number="121512042344",
            last_report_date=1618082952,
            last_report_watts=205,
            max_report_watts=253,
        ),
        "121512043086": EnvoyInverter(
            serial_number="121512043086",
            last_report_date=1618082942,
            last_report_watts=202,
            max_report_watts=250,
        ),
        "121512043093": EnvoyInverter(
            serial_number="121512043093",
            last_report_date=1618082928,
            last_report_watts=208,
            max_report_watts=255,
        ),
        "121512043135": EnvoyInverter(
            serial_number="121512043135",
            last_report_date=1618082923,
            last_report_watts=205,
            max_report_watts=254,
        ),
        "121512043153": EnvoyInverter(
            serial_number="121512043153",
            last_report_date=1618082935,
            last_report_watts=18,
            max_report_watts=146,
        ),
        "121512043173": EnvoyInverter(
            serial_number="121512043173",
            last_report_date=1618082966,
            last_report_watts=200,
            max_report_watts=247,
        ),
        "121512043200": EnvoyInverter(
            serial_number="121512043200",
            last_report_date=1618082955,
            last_report_watts=203,
            max_report_watts=253,
        ),
        "121512043222": EnvoyInverter(
            serial_number="121512043222",
            last_report_date=1618082957,
            last_report_watts=207,
            max_report_watts=254,
        ),
        "121512043574": EnvoyInverter(
            serial_number="121512043574",
            last_report_date=1618082936,
            last_report_watts=203,
            max_report_watts=253,
        ),
        "121512043587": EnvoyInverter(
            serial_number="121512043587",
            last_report_date=1618082934,
            last_report_watts=202,
            max_report_watts=253,
        ),
        "121512044424": EnvoyInverter(
            serial_number="121512044424",
            last_report_date=1618082954,
            last_report_watts=106,
            max_report_watts=239,
        ),
    }


@pytest.mark.asyncio
async def test_with_3_17_3_firmware_zero_production(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 3.17.3 firmware."""
    version = "3.17.3"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # Get envoy and let it probe with good data
    envoy = await get_mock_envoy(test_client_session, update=True)

    # Now override the production endpoint to return zeros for the next update
    override_mock(
        mock_aioresponse,
        "get",
        "http://127.0.0.1/api/v1/production",
        status=200,
        payload={
            "wattHoursToday": 0,
            "wattHoursSevenDays": 0,
            "wattHoursLifetime": 0,
            "wattsNow": 0,
        },
        repeat=True,
    )

    # Now update should fail with poor data quality
    with pytest.raises(EnvoyPoorDataQuality):
        await envoy.update()



"""Pyenphase Ensemble updater class."""

import logging
from typing import Any

from ..const import (
    ENSEMBLE_MIN_VERSION,
    URL_DRY_CONTACT_SETTINGS,
    URL_DRY_CONTACT_STATUS,
    URL_ENCHARGE_BATTERY,
    URL_ENSEMBLE_INVENTORY,
    URL_ENSEMBLE_SECCTRL,
    SupportedFeatures,
)
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.acb import EnvoyBatteryAggregate
from ..models.c6combiner import EnvoyC6CC
from ..models.collar import EnvoyCollar
from ..models.dry_contacts import EnvoyDryContactSettings, EnvoyDryContactStatus
from ..models.encharge import EnvoyEncharge, EnvoyEnchargeAggregate, EnvoyEnchargePower
from ..models.enpower import EnvoyEnpower
from ..models.envoy import EnvoyData
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyEnembleUpdater(EnvoyUpdater):
    """Class to handle updates for Ensemble devices."""

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this updater and return SupportedFeatures."""
        if self._envoy_version < ENSEMBLE_MIN_VERSION:
            _LOGGER.debug("Firmware too old for Ensemble support")
            return None

        # Check for various Ensemble support
        # The Ensemble Inventory endpoint will tell us if we have Enpower or Encharge support
        try:
            result = await self._json_probe_request(URL_ENSEMBLE_INVENTORY)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Ensemble Inventory endpoint not found: %s", e)
        except EnvoyAuthenticationRequired as e:
            # For URL_ENSEMBLE_INVENTORY some early V7 versions
            # return 401 when using aiohttp.
            _LOGGER.debug(
                "Skipping Ensemble Inventory endpoint as user does"
                " not have access to %s: %s",
                URL_ENSEMBLE_INVENTORY,
                e,
            )
            return None
        else:
            if not result or "error" in result:
                # Newer firmware with no Ensemble devices returns an empty list
                _LOGGER.debug("No Ensemble devices found")
                return None

            for item in result:
                if item["type"] == "ENPOWER":
                    self._supported_features |= SupportedFeatures.ENPOWER
                if item["type"] == "ENCHARGE":
                    self._supported_features |= SupportedFeatures.ENCHARGE
                if item["type"] == "COLLAR":
                    self._supported_features |= SupportedFeatures.COLLAR
                if item["type"] == "C6 COMBINER CONTROLLER":
                    self._supported_features |= SupportedFeatures.C6CC

        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the Envoy for this updater."""
        # Update Enpower and Encharge data if supported
        supported_features = self._supported_features
        ensemble_inventory_data: list[dict[str, Any]] = await self._json_request(
            URL_ENSEMBLE_INVENTORY
        )
        envoy_data.raw[URL_ENSEMBLE_INVENTORY] = ensemble_inventory_data

        ensemble_secctrl_data: dict[str, Any] = await self._json_request(
            URL_ENSEMBLE_SECCTRL
        )
        envoy_data.raw[URL_ENSEMBLE_SECCTRL] = await self._json_request(
            URL_ENSEMBLE_SECCTRL
        )

        if supported_features & SupportedFeatures.ENCHARGE:
            encharge_power_data: dict[str, Any] = await self._json_request(
                URL_ENCHARGE_BATTERY
            )
            envoy_data.raw[URL_ENCHARGE_BATTERY] = encharge_power_data
            power: dict[str, Any] = {
                device["serial_num"]: device
                for device in encharge_power_data["devices:"]
            }
            inventory: dict[str, Any] = {}
            for item in ensemble_inventory_data:
                if item["type"] != "ENCHARGE":
                    continue
                inventory = {device["serial_num"]: device for device in item["devices"]}

            envoy_data.encharge_inventory = {
                serial: EnvoyEncharge.from_api(inventory[serial])
                for serial in inventory
            }
            envoy_data.encharge_power = {
                serial: EnvoyEnchargePower.from_api(power[serial]) for serial in power
            }
            envoy_data.encharge_aggregate = EnvoyEnchargeAggregate.from_api(
                ensemble_secctrl_data
            )

        if supported_features & SupportedFeatures.ENPOWER:
            # Update Enpower data
            for item in ensemble_inventory_data:
                if item["type"] != "ENPOWER":
                    continue
                enpower_data = item["devices"][0]
            envoy_data.enpower = EnvoyEnpower.from_api(enpower_data)

            # Update dry contact data
            dry_contact_status_data: dict[str, Any] = await self._json_request(
                URL_DRY_CONTACT_STATUS
            )
            envoy_data.raw[URL_DRY_CONTACT_STATUS] = dry_contact_status_data

            dry_contact_settings_data: dict[str, Any] = await self._json_request(
                URL_DRY_CONTACT_SETTINGS
            )
            envoy_data.raw[URL_DRY_CONTACT_SETTINGS] = dry_contact_settings_data

            envoy_data.dry_contact_status = {
                relay["id"]: EnvoyDryContactStatus.from_api(relay)
                for relay in dry_contact_status_data["dry_contacts"]
            }
            envoy_data.dry_contact_settings = {
                relay["id"]: EnvoyDryContactSettings.from_api(relay)
                for relay in dry_contact_settings_data["dry_contacts"]
            }

        # production updater will set common_properties.ACB_batteries_reported
        # to count of ACB batteries. Check if production report found
        # acb batteries and if so report combined soc and max capacity
        # for Encharge and ACB batteries
        if self._common_properties.acb_batteries_reported:
            envoy_data.battery_aggregate = EnvoyBatteryAggregate.from_api(
                ensemble_secctrl_data
            )

        # IQ Meter collar seems like a single instance only
        if supported_features & SupportedFeatures.COLLAR:
            # Update Collar data
            for item in ensemble_inventory_data:
                if item["type"] != "COLLAR":
                    continue
                if item.get("devices"):
                    collar_data = item["devices"][0]
                    envoy_data.collar = EnvoyCollar.from_api(collar_data)

        # C6 Combiner seems like a single instance only
        if supported_features & SupportedFeatures.C6CC:
            # Update C6CC data
            for item in ensemble_inventory_data:
                if item["type"] != "C6 COMBINER CONTROLLER":
                    continue
                if item.get("devices"):
                    c6cc_data = item["devices"][0]
                    envoy_data.c6cc = EnvoyC6CC.from_api(c6cc_data)



"""JSON utilities for pyenphase."""

import logging
from typing import Any

import orjson

_LOGGER = logging.getLogger(__name__)


def json_loads(end_point: str, json_source: bytes | str) -> Any:
    """
    Deserialize a JSON string into a Python object

    :param end_point: source for json, used for debug log, typically endpoint on Envoy.
    :param json_source: json string, typically from request response content to Envoy.
    :return: deserialized JSON
    """
    try:
        return orjson.loads(json_source)
    except orjson.JSONDecodeError as e:
        _LOGGER.debug(
            "Unable to decode response from Envoy endpoint %s: %s\nResponse content: %s",
            end_point,
            e,
            json_source[:500] if isinstance(json_source, (bytes, str)) else json_source,
        )
        raise



"""Model for ENphase Envoy home data"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


def _find_active_interface(
    interfaces: list[dict[str, Any]], name: str
) -> dict[str, Any] | None:
    """Find an interface by interface name."""
    if interfaces:
        for interface in interfaces:
            if interface.get("interface") == name:
                return interface
    return None


@dataclass(slots=True)
class EnvoyInterfaceInformation:
    """Envoy Interface information data model."""

    primary_interface: str  #: name of primary (active) interface
    mac: str  #: mac of primary interface, "unknown" if missing
    interface_type: str  #: primary interface type, "unknown" if missing
    dhcp: bool  #: interfaces uses DHCP, False if missing
    software_build_epoch: int  #: envoy software build time, 0 if missing
    timezone: str  #: Timezone set in Envoy, "unknown" if missing

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> EnvoyInterfaceInformation | None:
        """
        Return active interface information configured in Envoy

        Parses the received JSON into EnvoyInterfaceInformation model data
        Source data must be sourced from URL_HOME.

        software_build_epoch, timezone are returned as is.
        network.primary_interface is returned as is, and used
        to find interface data in network.interfaces from which
        type, mac and dhcp are returned.

        Not all Envoy firmware version may return all data. Defaults for
        str members is unknown, int 0 and bool False

        Example json returned from /home endpoint:

            .. code-block:: json

                {
                    "software_build_epoch": 1719503966,
                    "timezone": "Europe/Amsterdam",
                    "current_date": "04/24/2025",
                    "current_time": "14:53",
                    "network": {
                        "web_comm": true,
                        "ever_reported_to_enlighten": true,
                        "last_enlighten_report_time": 1745499043,
                        "primary_interface": "eth0",
                        "interfaces": [
                            {
                                "type": "ethernet",
                                "interface": "eth0",
                                "mac": "00:1D:C0:7F:B6:3B",
                                "dhcp": true,
                                "ip": "192.168.3.112",
                                "signal_strength": 1,
                                "signal_strength_max": 1,
                                "carrier": true
                            },
                            {
                                "signal_strength": 0,
                                "signal_strength_max": 0,
                                "type": "wifi",
                                "interface": "wlan0",
                                "mac": "60:E8:5B:AB:9D:64",
                                "dhcp": true,
                                "ip": null,
                                "carrier": false,
                                "supported": true,
                                "present": true,
                                "configured": false,
                                "status": "connecting"
                            }
                        ]
                    }
                }

        :param data: json returned by /home endpoint
        :return: Envoy interface configuration information
        """
        # not sure if all firmware versions have all the needed information
        if not (network := data.get("network")):
            return None
        if not (
            interface := _find_active_interface(
                network.get("interfaces"), (name := network.get("primary_interface"))
            )
        ):
            return None
        return cls(
            primary_interface=name,
            mac=interface.get("mac", "unknown"),
            interface_type=interface.get("type", "unknown"),
            dhcp=interface.get("dhcp", False),
            software_build_epoch=data.get("software_build_epoch", 0),
            timezone=data.get("timezone", "unknown"),
        )



import logging
from unittest.mock import patch

import aiohttp
import pytest
import pytest_asyncio
from aioresponses import aioresponses
from syrupy import SnapshotAssertion

from pyenphase.ssl import NO_VERIFY_SSL_CONTEXT
from tests.syrupy import EnphaseSnapshotExtension


@pytest.fixture
def mock_aioresponse():
    """Return aioresponses fixture."""
    # Note: aioresponses will mock all ClientSession instances by default
    with aioresponses(passthrough=["http://127.0.0.1:8123"]) as m:
        yield m


@pytest_asyncio.fixture
async def test_client_session():
    """Create an aiohttp ClientSession with low timeout for tests."""
    timeout = aiohttp.ClientTimeout(total=5.0, connect=1.0, sock_read=1.0)
    connector = aiohttp.TCPConnector(ssl=NO_VERIFY_SSL_CONTEXT)
    session = aiohttp.ClientSession(timeout=timeout, connector=connector)
    yield session
    await session.close()


@pytest.fixture
def snapshot(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """Return snapshot assertion fixture with the Enphase extension."""
    return snapshot.use_extension(EnphaseSnapshotExtension)


@pytest.fixture(autouse=True)
def fast_tenacity():
    """Make tenacity retries fast by mocking time.sleep."""
    with patch("tenacity.nap.time"), patch("asyncio.sleep", return_value=None):
        yield


@pytest.fixture(autouse=True)
def setup_logging():
    """Set up logging for all tests."""
    logging.getLogger("pyenphase").setLevel(logging.DEBUG)



"""Test envoy home endpoint"""

import asyncio
import logging

import aiohttp
import pytest
from aioresponses import aioresponses

from pyenphase import Envoy
from pyenphase.models.home import EnvoyInterfaceInformation

from .common import (
    get_mock_envoy,
    load_fixture,
    load_json_fixture,
    override_mock,
    prep_envoy,
    start_7_firmware_mock,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_home_from_api_with_7_6_175() -> None:
    """Test home data from api"""
    # start with regular data first
    version = "7.6.175"

    # load mock data for home endpoint
    home_json = await load_json_fixture(version, "home")

    # test from_api method with eth0 interface
    home_data: EnvoyInterfaceInformation | None = EnvoyInterfaceInformation.from_api(
        home_json
    )

    # verify common data
    assert home_data
    assert home_data.software_build_epoch == 1719503966
    assert home_data.dhcp
    assert home_data.timezone == "Europe/Amsterdam"

    # verify interface data
    assert home_data.mac == "00:1D:C0:7F:B6:3B"
    assert home_data.primary_interface == "eth0"
    assert home_data.interface_type == "ethernet"

    # force wifi interface
    home_json["network"]["primary_interface"] = "wlan0"
    home_data = EnvoyInterfaceInformation.from_api(home_json)

    # verify interface data
    assert home_data
    assert home_data.mac == "60:E8:5B:AB:9D:64"
    assert home_data.primary_interface == "wlan0"
    assert home_data.interface_type == "wifi"

    # test missing interface key
    home_json["network"]["primary_interface"] = "missing"
    home_data = EnvoyInterfaceInformation.from_api(home_json)
    assert home_data is None

    # test handling missing interfaces part
    del home_json["network"]["interfaces"]
    home_data = EnvoyInterfaceInformation.from_api(home_json)
    assert not home_data

    # test handling missing network part
    del home_json["network"]
    home_data = EnvoyInterfaceInformation.from_api(home_json)
    assert not home_data


@pytest.mark.asyncio
async def test_interface_settings_with_7_6_175(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test home interface information data"""
    # start with regular data first
    version = "7.6.175"

    # Use prep_envoy to set up all required mocks
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # Create envoy using get_mock_envoy which handles all the setup
    envoy = await get_mock_envoy(test_client_session, update=False)

    # test interface_settings method
    home_data: EnvoyInterfaceInformation | None = await envoy.interface_settings()

    # validate common data
    assert home_data
    assert home_data.software_build_epoch == 1719503966
    assert home_data.dhcp
    assert home_data.timezone == "Europe/Amsterdam"

    # validate interface data
    assert home_data.mac == "00:1D:C0:7F:B6:3B"
    assert home_data.primary_interface == "eth0"
    assert home_data.interface_type == "ethernet"

    # load mock data for home endpoint
    home_json = await load_json_fixture(version, "home")
    # Change mock to use wlan interface
    home_json["network"]["primary_interface"] = "wlan0"
    # and mock new data
    override_mock(
        mock_aioresponse, "get", "https://127.0.0.1/home", status=200, payload=home_json
    )

    # get interface data, subsequent call data is returned from cache
    home_data = await envoy.interface_settings()

    # interface data should come from cache and still be the same
    assert home_data
    assert home_data.mac == "00:1D:C0:7F:B6:3B"
    assert home_data.primary_interface == "eth0"
    assert home_data.interface_type == "ethernet"

    # call setup to invalidate cached interface data
    await envoy.setup()

    # now interface data should be reflect latest mocked data and reflect wlan interface
    home_data = await envoy.interface_settings()
    assert home_data
    assert home_data.mac == "60:E8:5B:AB:9D:64"
    assert home_data.primary_interface == "wlan0"
    assert home_data.interface_type == "wifi"


@pytest.mark.asyncio
async def test_home_endpoint_errors_with_7_6_175(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test home interface information data"""
    caplog.set_level(logging.DEBUG)

    # start with regular data first
    version = "7.6.175"

    # Set up auth mocks
    start_7_firmware_mock(mock_aioresponse)

    # Set up info endpoint for setup
    info_data = await load_fixture(version, "info")
    mock_aioresponse.get(
        "https://127.0.0.1/info", status=200, body=info_data, repeat=True
    )

    # Create and setup envoy
    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")

    # test not-found error
    mock_aioresponse.get("https://127.0.0.1/home", status=404)
    await envoy.interface_settings()
    assert "Failure getting interface information" in caplog.text
    caplog.clear()

    # test server error
    mock_aioresponse.get("https://127.0.0.1/home", status=500)
    await envoy.interface_settings()
    assert "Failure getting interface information" in caplog.text
    caplog.clear()

    mock_aioresponse.get(
        "https://127.0.0.1/home", exception=aiohttp.ClientError("Test Networkerror")
    )
    await envoy.interface_settings()
    assert "Failure getting interface information" in caplog.text
    caplog.clear()

    mock_aioresponse.get(
        "https://127.0.0.1/home",
        exception=asyncio.TimeoutError("Test timeoutexception"),
    )
    await envoy.interface_settings()
    assert "Failure getting interface information" in caplog.text


@pytest.mark.asyncio
async def test_home_endpoint_non_json_response(
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test home endpoint returning non-JSON content (older Envoys)."""
    caplog.set_level(logging.DEBUG)

    version = "7.6.175"

    start_7_firmware_mock(mock_aioresponse)

    info_data = await load_fixture(version, "info")
    mock_aioresponse.get(
        "https://127.0.0.1/info", status=200, body=info_data, repeat=True
    )

    envoy = Envoy("127.0.0.1", client=test_client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")

    # Mock /home returning HTML (non-JSON) with 200 status
    mock_aioresponse.get(
        "https://127.0.0.1/home",
        status=200,
        body=b"<html><body>Not Found</body></html>",
    )
    result = await envoy.interface_settings()
    assert result is None
    assert "Failure getting interface information" in caplog.text



"""Model for the ACB Battery."""

# Data Source: URL_ENSEMBLE_SECCTRL (primary) & URL_PRODUCTION_JSON

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EnvoyBatteryAggregate:
    """Model for combined Encharge and ACB batteries aggregate data."""

    #: Sum of Encharge aggregate and ACB aggregate current battery energy content
    #: from ENC_agg_avail_energy and ACB_agg_energy.
    available_energy: int
    #: Combined State of charge for all Encharge and ACB batteries from agg_soc.
    state_of_charge: int
    #: Combined total maximum capacity for all Encharge and ACB batteries from Max_energy.
    max_available_capacity: int

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> EnvoyBatteryAggregate:
        """
        Fill Aggregated battery data from Envoy data format.

        Source data parts of URL_ENSEMBLE_SECCTRL
            .. code-block:: json

                {

                    "agg_soc": 39,
                    "Max_energy": 7220,

                    "ENC_agg_avail_energy": 350,

                    "Enc_max_available_capacity": 3500,
                    "ACB_agg_soc": 25,
                    "ACB_agg_energy": 930,

                }

        Args:
            data (dict[str, Any]): JSON returned from URL_ENSEMBLE_SECCTRL

        Returns:
            EnvoyBatteryAggregate: Aggregated Battery data for all Encharge and ACB batteries

        """
        return cls(
            available_energy=data["ENC_agg_avail_energy"] + data["ACB_agg_energy"],
            max_available_capacity=data["Max_energy"],
            state_of_charge=data["agg_soc"],
        )


@dataclass(slots=True)
class EnvoyACBPower:
    """Model for the ACB battery power."""

    #: Current discharge/charge power for ACB batteries from wNow.
    power: int
    #: Current available capacity in Wh for ACB batteries from whNow
    charge_wh: int
    #: Current SOC in percentage for ACB batteries from percentFull
    state_of_charge: int
    #: Current state for ACB batteries (discharging/idle/charging) from state
    state: str
    #: Number of reported ACB batteries from activeCount
    batteries: int

    @classmethod
    def from_production(
        cls, data: dict[str, Any], acb_segment: int = 0
    ) -> EnvoyACBPower:
        """
        Fill ACB battery power data from Envoy data format.

        Source data URL_PRODUCTION_JSON["storage"]
            .. code-block:: json

                "storage": [{
                    "type": "acb",
                    "activeCount": 3,
                    "readingTime": 1731943992,
                    "wNow": 260,
                    "whNow": 930,
                    "state": "discharging",
                    "percentFull": 25
                }]


        Args:
            data (dict[str, Any]): JSON returned from URL_PRODUCTION_JSON
            acb_segment (int): segment to process from storage list, default is 0

        Returns:
            EnvoyACBPower: ACB battery current power out/in and energy content and status

        """
        storage_data = data["storage"][acb_segment]
        return cls(
            power=storage_data["wNow"],
            charge_wh=storage_data["whNow"],
            state_of_charge=storage_data["percentFull"],
            state=storage_data["state"],
            batteries=storage_data["activeCount"],
        )



"""Model for the Encharge/IQ Battery."""

# Data Source: URL_ENSEMBLE_INVENTORY (primary) & URL_ENCHARGE_BATTERY

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EnvoyEnchargeAggregate:
    """Model for Encharge aggregate data."""

    available_energy: int
    backup_reserve: int
    state_of_charge: int
    reserve_state_of_charge: int
    configured_reserve_state_of_charge: int
    max_available_capacity: int

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> EnvoyEnchargeAggregate:
        """Initialize from the API."""
        return cls(
            available_energy=data["ENC_agg_avail_energy"],
            backup_reserve=data["ENC_agg_backup_energy"],
            state_of_charge=data["ENC_agg_soc"],
            reserve_state_of_charge=data["adjusted_backup_soc"],
            configured_reserve_state_of_charge=data["configured_backup_soc"],
            max_available_capacity=data["Enc_max_available_capacity"],
        )


@dataclass(slots=True)
class EnvoyEnchargePower:
    """Model for the Encharge/IQ battery power."""

    apparent_power_mva: int
    real_power_mw: int
    soc: int

    @classmethod
    def from_api(cls, power: dict[str, Any]) -> EnvoyEnchargePower:
        """Initialize from the API."""
        return cls(
            apparent_power_mva=power["apparent_power_mva"],
            real_power_mw=power["real_power_mw"],
            soc=power["soc"],
        )


@dataclass(slots=True)
class EnvoyEncharge:
    """Model for the Encharge/IQ battery."""

    admin_state: int
    admin_state_str: str
    bmu_firmware_version: str
    comm_level_2_4_ghz: int
    comm_level_sub_ghz: int
    communicating: bool
    dc_switch_off: bool
    encharge_capacity: int
    encharge_revision: int
    firmware_loaded_date: int
    firmware_version: str
    installed_date: int
    last_report_date: int
    led_status: int
    max_cell_temp: int
    operating: bool | None
    part_number: str
    percent_full: int
    serial_number: str
    temperature: int
    temperature_unit: str
    zigbee_dongle_fw_version: str | None

    @classmethod
    def from_api(cls, inventory: dict[str, Any]) -> EnvoyEncharge:
        """Initialize from the API."""
        return cls(
            admin_state=inventory["admin_state"],
            admin_state_str=inventory["admin_state_str"],
            bmu_firmware_version=inventory["bmu_fw_version"],
            comm_level_2_4_ghz=inventory["comm_level_2_4_ghz"],
            comm_level_sub_ghz=inventory["comm_level_sub_ghz"],
            communicating=inventory["communicating"],
            dc_switch_off=inventory["dc_switch_off"],
            encharge_capacity=inventory["encharge_capacity"],
            encharge_revision=inventory["encharge_rev"],
            firmware_loaded_date=inventory["img_load_date"],
            firmware_version=inventory["img_pnum_running"],
            installed_date=inventory["installed"],
            last_report_date=inventory["last_rpt_date"],
            led_status=inventory["led_status"],
            max_cell_temp=inventory["maxCellTemp"],
            operating=inventory.get("operating"),  # Firmware 8+ does not have this key
            part_number=inventory["part_num"],
            percent_full=inventory["percentFull"],
            serial_number=inventory["serial_num"],
            temperature=inventory["temperature"],
            temperature_unit="C",
            zigbee_dongle_fw_version=inventory.get(
                "zigbee_dongle_fw_version"
            ),  # Firmware 8+ does not have this key
        )



import asyncio
import logging
import os
from pprint import pprint

from aiohttp import ClientResponse

from pyenphase.envoy import Envoy

logging.basicConfig(level=logging.DEBUG)


async def main() -> None:
    envoy = Envoy(os.environ.get("ENVOY_HOST", "envoy.local"))

    await envoy.setup()

    username = os.environ.get("ENVOY_USERNAME")
    password = os.environ.get("ENVOY_PASSWORD")
    token = os.environ.get("ENVOY_TOKEN")

    await envoy.authenticate(username=username, password=password, token=token)

    # Test https://enphase.com/download/iq-gateway-access-using-local-apis-or-local-ui-token-based-authentication-tech-brief endpoints

    print(await envoy.update())

    end_points = [
        "/ivp/livedata/status",
        "/api/v1/production",
        "/api/v1/production/inverters",
        "/production.json",
        "/production",
        "/ivp/meters",
        "/ivp/meters/readings",
        "/ivp/meters/reports/consumption",
        "/ivp/ensemble/inventory",
        "/ivp/ensemble/dry_contacts",
        "/ivp/ss/dry_contact_settings",
        "/ivp/pdm/device_data",
    ]

    for end_point in end_points:
        try:
            response: ClientResponse = await envoy.request(end_point)
        except Exception as e:
            print(e)
            continue
        json_dict = await response.json()
        print((end_point, "=" * 80))
        pprint(json_dict)
        print((end_point, "=" * 80))


asyncio.run(main())



"""Model for common properties of an envoy."""

from dataclasses import dataclass, field

from ..models.meters import EnvoyPhaseMode


@dataclass(slots=True)
class CommonProperties:
    """
    Model for common properties for EnvoyUpdater class updaters.

    Class :class:`EnvoyUpdater` implementations
    each collect a specific data set or information that can be provided by
    an Envoy. Depending on the actual model, firmware, and installed
    an Envoy may need one or more updaters to provide all data
    components, for :any:`pyenphase.Envoy.update`.

    Although each updater has its specific scope, some may need to share
    information or make operational information available. One set of
    properties are used during probe to share amongst updaters and with
    client applications. These should be reset at each probe run. A second
    set of properties are specific for an updater during its runtime. The
    updater is in control of resetting the property as needed.
    """

    # probe properties here, also add to reset_probe_properties
    # shared amongst production updaters, needs reset before probe
    #: Probe property, Fallback production endpoints for Metered without CT.
    #: Solar production data has to be collected from different
    #: sources in the Envoy, depending on actual Envoy configuration,
    #: each utilizing a different updater. This list will be filled
    #: with candidate updaters to use.
    production_fallback_list: list[str] = field(
        default_factory=list
    )  #: Fallback production endpoints for Metered without CT

    #: ACB batteries report current power in production and in the Ensemble SECCTRL endpoint
    #: The Ensemble updater should only report combined ACB and Encharge if production reported data
    acb_batteries_reported: int = 0

    #: imeter flag from /info. If true envoy is metered type
    #: used to detect metered without actual CT installed to enable picking correct data
    imeter_info: bool = False

    # other properties from here, reset by originator

    # controlled by meters updater
    #: meters updater, number of phases configured in envoy
    phase_count: int = 0
    #: meters updater, number of active ct meters
    ct_meter_count: int = 0
    #: meters updater, phase mode configured in the CT meters
    phase_mode: EnvoyPhaseMode | None = None
    #: meters updater, list of installed meter types, if installed
    meter_types: list[str] = field(default_factory=list)
    #: production updater, number of phases actually reporting phase data
    active_phase_count: int = 0

    def reset_probe_properties(self, is_metered: bool = False) -> None:
        """
        Reset common properties at start of probe.

        Probe common properties are reset at each probe by :any:`Envoy.probe`
        to avoid sticking values. This should only be done for common properties
        shared among updaters. Any common properties set outside of probe or controlled
        by a specific updater, should be reset at different moments by the
        owner of the property.

        Shared common properties to reset:

            - production_fallback_list shared amongst production updaters
            - ACB_batteries_reported shared between production and Ensemble
            - imeter_info setting from /info indicating envoy is metered type

        :return: None
        """
        # shared amongst production updaters
        self.production_fallback_list = []
        self.imeter_info = is_metered

        # shared between production and ensemble
        self.acb_batteries_reported = 0



"""Model for the Enpower/IQ System Controller."""

# Data Source: URL_ENSEMBLE_INVENTORY
from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EnvoyEnpower:
    """Model for the Enpower/IQ System Controller."""

    grid_mode: str
    admin_state: int
    admin_state_str: str
    comm_level_2_4_ghz: int
    comm_level_sub_ghz: int
    communicating: bool
    firmware_loaded_date: int
    firmware_version: str
    installed_date: int
    last_report_date: int
    mains_admin_state: str
    mains_oper_state: str
    operating: bool | None
    part_number: str
    serial_number: str
    temperature: int
    temperature_unit: str
    zigbee_dongle_fw_version: str | None

    @classmethod
    def from_api(
        cls,
        enpower: dict[str, Any],
    ) -> EnvoyEnpower:
        """Initialize from the API."""
        return cls(
            grid_mode=enpower["Enpwr_grid_mode"],
            admin_state=enpower["admin_state"],
            admin_state_str=enpower["admin_state_str"],
            comm_level_2_4_ghz=enpower["comm_level_2_4_ghz"],
            comm_level_sub_ghz=enpower["comm_level_sub_ghz"],
            communicating=enpower["communicating"],
            firmware_loaded_date=enpower["img_load_date"],
            firmware_version=enpower["img_pnum_running"],
            installed_date=enpower["installed"],
            last_report_date=enpower["last_rpt_date"],
            mains_admin_state=enpower["mains_admin_state"],
            mains_oper_state=enpower["mains_oper_state"],
            operating=enpower.get("operating"),  # Firmware 8+ does not have this field
            part_number=enpower["part_num"],
            serial_number=enpower["serial_num"],
            temperature=enpower["temperature"],
            temperature_unit="F",
            zigbee_dongle_fw_version=enpower.get(
                "zigbee_dongle_fw_version"
            ),  # Firmware 8+ does not have this field
        )



from dataclasses import asdict, is_dataclass
from typing import Any

from syrupy.extensions import AmberSnapshotExtension
from syrupy.extensions.amber import AmberDataSerializer
from syrupy.types import PropertyFilter, PropertyMatcher, PropertyPath, SerializableData


class EnphaseSnapshotSerializer(AmberDataSerializer):
    """
    Enphase snapshot serializer for Syrupy.

    Handles special cases for Enphase data structures.
    """

    @classmethod
    def _serialize(
        cls,
        data: SerializableData,
        *,
        depth: int = 0,
        exclude: PropertyFilter | None = None,
        include: PropertyFilter | None = None,
        matcher: PropertyMatcher | None = None,
        path: PropertyPath = (),
        visited: set[Any] | None = None,
    ) -> str:
        """
        Pre-process data before serializing.

        This allows us to handle specific cases for Enphase data structures.
        """
        serializable_data = data
        if is_dataclass(type(data)):
            serializable_data = asdict(data)

        return super()._serialize(
            serializable_data,
            depth=depth,
            exclude=exclude,
            include=include,
            matcher=matcher,
            path=path,
            visited=visited,
        )


class EnphaseSnapshotExtension(AmberSnapshotExtension):
    """Enphase extension for Syrupy."""

    VERSION = "1"
    """Current version of serialization format.

    Need to be bumped when we change the EnphasenapshotSerializer.
    """

    serializer_class: type[AmberDataSerializer] = EnphaseSnapshotSerializer



"""Model for the Envoy's CT Meters."""

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any, TypedDict


class EnvoyPhaseMode(StrEnum):
    SPLIT = "split"
    THREE = "three"
    SINGLE = "single"


class CtType(StrEnum):
    #: Solar production to main panel or Combiner/System controller
    PRODUCTION = "production"
    #: Net consumption between main panel and grid
    NET_CONSUMPTION = "net-consumption"
    #: Total consumption between main panel and house load
    TOTAL_CONSUMPTION = "total-consumption"
    #: Power and energy between battery and main panel or Combiner/System controller
    STORAGE = "storage"
    #: Power and energy backfeed between Combiner/System controller and main panel
    BACKFEED = "backfeed"
    #: Power and energy assumed between Combiner/System controller and backup load
    LOAD = "load"
    #: Power and Energy between Combiner/System controller and EV charger
    EVSE = "evse"
    #: Power and energy between 3rd-party solar and Combiner/System controller
    PV3P = "pv3p"


class CtState(StrEnum):
    ENABLED = "enabled"
    DISABLED = "disabled"


class CtMeterStatus(StrEnum):
    NORMAL = "normal"
    NOT_METERING = "not-metering"
    CHECK_WIRING = "check-wiring"


class CtStatusFlags(StrEnum):
    PODUCTION_IMBALANCE = "production-imbalance"
    NEGATIVE_PRODUCTION = "negative-production"
    POWER_ON_UNUSED_PHASE = "power-on-unused-phase"
    NEGATIVE_TOTAL_CONSUMPTION = "negative-total-consumption"


class CtMeterData(TypedDict):
    eid: str
    state: CtState
    measurementType: CtType
    phaseMode: EnvoyPhaseMode
    phaseCount: int
    meteringStatus: CtMeterStatus
    statusFlags: list[CtStatusFlags]


@dataclass(slots=True)
class EnvoyMeterData:
    """Model for the Envoy's CT meter data."""

    eid: str  #: CT meter identifier
    timestamp: int  #: Time of measurement
    energy_delivered: int  #: Lifetime Energy delivered through CT
    energy_received: int  #: Lifetime Energy received through CT
    active_power: int  #: Current power exchange through CT, positive is delivering, negative is receiving
    power_factor: float  #: Power factor reported for CT measurement
    voltage: float  #: Voltage on circuit, when multiphase sum of voltage of individual phases
    current: float  #: Current measured by CT
    frequency: float  #: frequency measured by CT
    state: CtState | None  #: Actual State of CT
    measurement_type: CtType | None  #: Measurement type configured for CT
    metering_status: CtMeterStatus | None  #: CT Measurement status
    status_flags: list[CtStatusFlags] | None  #: CT status flags.

    @classmethod
    def from_api(
        cls, data: dict[str, Any], meter_status: CtMeterData
    ) -> EnvoyMeterData:
        """Return CT meter data from /ivp/meters and ivp/meters/reading json."""
        return cls(
            eid=data["eid"],
            timestamp=data["timestamp"],
            energy_delivered=round(data["actEnergyDlvd"]),
            energy_received=round(data["actEnergyRcvd"]),
            active_power=round(data["activePower"]),
            power_factor=data["pwrFactor"],
            voltage=data["voltage"],
            current=data["current"],
            frequency=data["freq"],
            state=meter_status["state"],
            measurement_type=meter_status["measurementType"],
            metering_status=meter_status["meteringStatus"],
            # statusFlags are not present in older firmware versions
            status_flags=meter_status.get("statusFlags"),
        )

    @classmethod
    def from_phase(
        cls, data: dict[str, Any], meter_status: CtMeterData, phase: int
    ) -> EnvoyMeterData | None:
        """Return CT meter phase data from /ivp/meters and ivp/meters/reading json."""
        if "channels" not in data:
            return None
        # phase data is in channels list
        channels = data["channels"]
        if len(channels) <= phase:
            return None

        return cls.from_api(channels[phase], meter_status)



"""Pyenphase SSL helper"""

import contextlib
import ssl


def create_no_verify_ssl_context() -> ssl.SSLContext:
    """
    Return an SSL context that does not verify the server certificate.

    This is a copy of aiohttp's create_default_context() function, with the
    ssl verify turned off and old SSL versions enabled.

    https://github.com/aio-libs/aiohttp/blob/33953f110e97eecc707e1402daa8d543f38a189b/aiohttp/connector.py#L911

    :return: SSLcontext with ssl verify turned off.
    """
    sslcontext = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    sslcontext.check_hostname = False
    sslcontext.verify_mode = ssl.CERT_NONE
    # Allow all ciphers rather than only Python 3.10 default
    sslcontext.set_ciphers("DEFAULT")
    with contextlib.suppress(AttributeError):
        # This only works for OpenSSL >= 1.0.0
        sslcontext.options |= ssl.OP_NO_COMPRESSION
    sslcontext.set_default_verify_paths()
    return sslcontext


#: Alias for :any:`create_no_verify_ssl_context`
#:
#: .. code-block:: python
#:
#:     import aiohttp
#:     from pyenphase.ssl import NO_VERIFY_SSL_CONTEXT
#:
#:     connector = aiohttp.TCPConnector(ssl=NO_VERIFY_SSL_CONTEXT)
#:     client = aiohttp.ClientSession(connector=connector)
NO_VERIFY_SSL_CONTEXT = create_no_verify_ssl_context()


def create_default_ssl_context() -> ssl.SSLContext:
    """Create aiohttp client with default SSL context."""
    return ssl.create_default_context()


#: Alias for :any:`create_default_ssl_context`
#:
#: .. code-block:: python
#:
#:    import aiohttp
#:    from pyenphase.ssl import SSL_CONTEXT
#:
#:    connector = aiohttp.TCPConnector(ssl=SSL_CONTEXT)
#:    async with aiohttp.ClientSession(connector=connector) as client:
#:        response = await client.post(url, json=json, data=data)
#:
SSL_CONTEXT = create_default_ssl_context()



"""Model for the Envoy's consumption data."""

# Data Source: URL_PRODUCTION

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EnvoySystemConsumption:
    """Model for the Envoy's (total, house) consumption data."""

    watt_hours_lifetime: int  #: Lifetime Energy consumed (total-consumption, house)
    watt_hours_last_7_days: (
        int  #: Energy consumed in previous 7 days (not including today)
    )
    watt_hours_today: (
        int  #: Energy consumption since start of day (total-consumption, house)
    )
    watts_now: int  #: Current Power consumption (total-consumption, house)

    @classmethod
    def from_production(
        cls, data: dict[str, Any], consumption_segment: int = 0
    ) -> EnvoySystemConsumption:
        """
        Initialize from the production API.

        :param data: JSON reply from /production endpoint
        :return: Lifetime, last 7 days, todays energy and current power for total-consumption
        """
        consumption = data["consumption"][consumption_segment]
        return cls(
            watt_hours_lifetime=round(consumption["whLifetime"]),
            watt_hours_last_7_days=round(consumption["whLastSevenDays"]),
            watt_hours_today=round(consumption["whToday"]),
            watts_now=round(consumption["wNow"]),
        )

    @classmethod
    def from_production_phase(
        cls, data: dict[str, Any], phase: int, consumption_segment: int = 0
    ) -> EnvoySystemConsumption | None:
        """
        Initialize from the production API phase data.

        :param data: JSON reply from /production endpoint
        :param phase: Index (0-2) in [lines] segment for which to return data
        :return: Lifetime, last 7 days, todays energy and current power for total-consumption phase
        """
        # get first consumtpion section which is the total-consumption one.
        consumption = data["consumption"][consumption_segment]
        phases = consumption.get("lines")

        # Only return data if phase is present.
        if not phases or phase >= len(phases):
            return None

        phase_data = phases[phase]
        return cls(
            watt_hours_lifetime=round(phase_data["whLifetime"]),
            watt_hours_last_7_days=round(phase_data["whLastSevenDays"]),
            watt_hours_today=round(phase_data["whToday"]),
            watts_now=round(phase_data["wNow"]),
        )



"""common test functions for pyenphase."""

import asyncio
import json
import logging
from collections.abc import Generator
from contextlib import contextmanager
from os import listdir
from os.path import isfile, join
from pathlib import Path
from typing import Any

import aiohttp
import orjson
from aioresponses import aioresponses
from awesomeversion import AwesomeVersion

from pyenphase import AUTH_TOKEN_MIN_VERSION, Envoy
from pyenphase.envoy import SupportedFeatures
from pyenphase.updaters.base import EnvoyUpdater


@contextmanager
def temporary_log_level(logger_name: str, level: int) -> Generator[None, None, None]:
    """Temporarily change the log level of a logger."""
    logger = logging.getLogger(logger_name)
    original_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(original_level)


def _fixtures_dir() -> Path:
    """Return location of fixture folder."""
    return Path(__file__).parent / "fixtures"


def _load_fixture(path: str) -> str:
    """Return fixture file content as text."""
    with open(path) as read_in:
        return read_in.read()


async def load_fixture(version: str, name: str) -> str:
    """Return fixture file content as text in executor."""
    path: str = f"{_fixtures_dir()}/{version}/{name}"
    content: str = await asyncio.get_running_loop().run_in_executor(
        None, _load_fixture, path
    )
    return content


async def load_json_fixture(version: str, name: str) -> dict[str, Any]:
    """Return fixture file content as JSON."""
    json: str = await load_fixture(version, name)
    return orjson.loads(json)


async def load_json_list_fixture(version: str, name: str) -> list[dict[str, Any]]:
    """Return list[json] fixture file content as json"""
    json: str = await load_fixture(version, name)
    return orjson.loads(json)


def _fixture_files(path: str) -> list[str]:
    """Get fixture files list"""
    return [f for f in listdir(path) if isfile(join(path, f))]


async def fixture_files(version: str) -> list[str]:
    """Get fixture files list in executor"""
    path: str = f"{_fixtures_dir()}/{version}"
    files: list[str] = await asyncio.get_running_loop().run_in_executor(
        None, _fixture_files, path
    )
    return files


def start_7_firmware_mock(mock_aioresponse: aioresponses) -> None:
    """Setup response mocks for Enlighten and Envoy token requests."""
    # Use repeat=True since auth might create its own session
    # Mock both with and without the trailing ?
    mock_aioresponse.post(
        "https://enlighten.enphaseenergy.com/login/login.json?",
        status=200,
        payload={
            "session_id": "1234567890",
            "user_id": "1234567890",
            "user_name": "test",
            "first_name": "Test",
            "is_consumer": True,
            "manager_token": "1234567890",
        },
        repeat=True,
    )
    mock_aioresponse.post(
        "https://enlighten.enphaseenergy.com/login/login.json",
        status=200,
        payload={
            "session_id": "1234567890",
            "user_id": "1234567890",
            "user_name": "test",
            "first_name": "Test",
            "is_consumer": True,
            "manager_token": "1234567890",
        },
        repeat=True,
    )
    mock_aioresponse.post(
        "https://entrez.enphaseenergy.com/tokens",
        status=200,
        body="token",
        repeat=True,
    )
    # Mock the JWT check endpoint on the Envoy
    mock_aioresponse.get("https://127.0.0.1/auth/check_jwt", status=200, repeat=True)


async def get_mock_envoy(
    client_session: aiohttp.ClientSession | None, update: bool = True
) -> Envoy:
    """Return a mock Envoy."""
    host = "127.0.0.1"
    envoy = Envoy(host, client=client_session)
    await envoy.setup()
    await envoy.authenticate("username", "password")
    if update:
        await envoy.update()
        await envoy.update()  # make sure we can update twice
    return envoy


def latest_request(
    mock_aioresponse: aioresponses, method: str, url: str
) -> tuple[int, bytes]:
    """Return count of matched request and last request data."""
    requests = [
        req
        for req in mock_aioresponse.requests
        if req[0] == method and url in str(req[1])
    ]
    if not requests:
        return 0, b""
    return len(requests), mock_aioresponse.requests[requests[-1]][-1].kwargs.get("data")


def override_mock(
    mock_aioresponse: aioresponses, method: str, url: str, **kwargs: Any
) -> None:
    """Override an existing mock by removing it first and adding a new one."""
    from yarl import URL

    url_obj = URL(url)

    # Remove existing mocks for this URL and method from _matches
    keys_to_remove = []
    for key, matcher in mock_aioresponse._matches.items():
        # Check if this matches our URL and method
        url_matches = False
        if hasattr(matcher.url_or_pattern, "match"):
            # It's a regex pattern
            url_matches = bool(matcher.url_or_pattern.match(str(url_obj)))
        else:
            # It's a URL
            url_matches = str(matcher.url_or_pattern).rstrip("/") == str(
                url_obj
            ).rstrip("/")

        if url_matches and matcher.method.lower() == method.lower():
            keys_to_remove.append(key)

    # Remove the matching mocks
    for key in keys_to_remove:
        del mock_aioresponse._matches[key]

    # Add the new mock
    getattr(mock_aioresponse, method.lower())(url, **kwargs)


def endpoint_path(version: str, host: str) -> str:
    return f"http{'s' if AwesomeVersion(version.split('_')[0]) >= AUTH_TOKEN_MIN_VERSION else ''}://{host}"


async def prep_envoy(
    mock_aioresponse: aioresponses,
    host: str,
    version: str,  #: name of version folder to read fixtures from
) -> list[str]:
    """Setup response mocks for envoy requests and return list of found mock files."""
    files: list[str] = await fixture_files(version)

    # Helper to create full URLs
    full_host = endpoint_path(version, host)

    def url(path: str) -> str:
        return f"{full_host}{path}"

    def url_https(path: str) -> str:
        return f"https://{host}{path}"

    def url_http(path: str) -> str:
        return f"http://{host}{path}"

    mock_aioresponse.get(
        url_http("/info"),
        status=200,
        body=await load_fixture(version, "info"),
        repeat=True,
    )
    mock_aioresponse.get(
        url_https("/info"),
        status=200,
        body=await load_fixture(version, "info"),
        repeat=True,
    )
    mock_aioresponse.get(url("/info.xml"), status=200, body="", repeat=True)

    if "ivp_meters" in files:
        try:
            mock_aioresponse.get(
                url("/ivp/meters"),
                status=200,
                payload=(await load_json_fixture(version, "ivp_meters")),
                repeat=True,
            )
        except json.decoder.JSONDecodeError:
            # v3 fw with html return 401
            mock_aioresponse.get(url("/ivp/meters"), status=401, repeat=True)
            # mock_aioresponse.get(url_http("/ivp/meters"), status=401, repeat=True)
    else:
        mock_aioresponse.get(url("/ivp/meters"), status=404, repeat=True)
        # mock_aioresponse.get(url_http("/ivp/meters"), status=404, repeat=True)

    if "ivp_meters_readings" in files:
        mock_aioresponse.get(
            url("/ivp/meters/readings"),
            status=200,
            payload=await load_json_fixture(version, "ivp_meters_readings"),
            repeat=True,
        )
    else:
        mock_aioresponse.get(url("/ivp/meters/readings"), status=404, repeat=True)
        # mock_aioresponse.get(url_http("/ivp/meters/readings"), status=404, repeat=True)

    if "production" in files:
        try:
            json_data = await load_json_fixture(version, "production")
            mock_aioresponse.get(
                url("/production"), status=200, payload=json_data, repeat=True
            )
        except json.decoder.JSONDecodeError:
            # v3 fw reports production in html format
            mock_aioresponse.get(
                url("/production"),
                status=200,
                body=await load_fixture(version, "production"),
                repeat=True,
            )
    else:
        mock_aioresponse.get(url("/production"), status=404, repeat=True)
        # mock_aioresponse.get(url_http("/production"), status=404, repeat=True)

    if "production.json" in files:
        try:
            json_data = await load_json_fixture(version, "production.json")
            mock_aioresponse.get(
                url("/production.json"), status=200, payload=json_data, repeat=True
            )
            mock_aioresponse.get(
                url("/production.json?details=1"),
                status=200,
                payload=json_data,
                repeat=True,
            )
        except json.decoder.JSONDecodeError:
            mock_aioresponse.get(url("/production.json"), status=404, repeat=True)
            # mock_aioresponse.get(url_http("/production.json"), status=404, repeat=True)
            mock_aioresponse.get(
                url("/production.json?details=1"), status=404, repeat=True
            )
    else:
        mock_aioresponse.get(url("/production.json"), status=404, repeat=True)

    if "api_v1_production" in files:
        # Check if this is a bad_auth version by looking at the fixture content
        api_v1_prod_data = await load_json_fixture(version, "api_v1_production")
        status = (
            401
            if "status" in api_v1_prod_data and api_v1_prod_data["status"] == 401
            else 200
        )

        mock_aioresponse.get(
            url("/api/v1/production"),
            status=status,
            payload=api_v1_prod_data,
            repeat=True,
        )
    else:
        mock_aioresponse.get(url("/api/v1/production"), status=404, repeat=True)

    if "api_v1_production_inverters" in files:
        # Check if this is a bad_auth version by looking at the fixture content
        api_v1_inv_data = await load_json_fixture(
            version, "api_v1_production_inverters"
        )
        status = (
            401
            if isinstance(api_v1_inv_data, dict)
            and "status" in api_v1_inv_data
            and api_v1_inv_data["status"] == 401
            else 200
        )

        mock_aioresponse.get(
            url("/api/v1/production/inverters"),
            status=status,
            payload=api_v1_inv_data,
            repeat=True,
        )
    else:
        mock_aioresponse.get(
            url("/api/v1/production/inverters"), status=404, repeat=True
        )

    if "ivp_ensemble_inventory" in files:
        mock_aioresponse.get(
            url("/ivp/ensemble/inventory"),
            status=200,
            payload=await load_json_fixture(version, "ivp_ensemble_inventory"),
            repeat=True,
        )
        mock_aioresponse.get(
            url_http("/ivp/ensemble/inventory"),
            status=200,
            payload=await load_json_fixture(version, "ivp_ensemble_inventory"),
            repeat=True,
        )
    else:
        mock_aioresponse.get(url("/ivp/ensemble/inventory"), status=404, repeat=True)
        mock_aioresponse.get(
            url_http("/ivp/ensemble/inventory"), status=404, repeat=True
        )

    if "ivp_ensemble_dry_contacts" in files:
        try:
            json_data = await load_json_fixture(version, "ivp_ensemble_dry_contacts")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/ivp/ensemble/dry_contacts"),
            status=200,
            payload=json_data,
            repeat=True,
        )
        mock_aioresponse.post(
            url("/ivp/ensemble/dry_contacts"),
            status=200,
            payload=json_data,
            repeat=True,
        )

    if "ivp_ss_dry_contact_settings" in files:
        try:
            json_data = await load_json_fixture(version, "ivp_ss_dry_contact_settings")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/ivp/ss/dry_contact_settings"),
            status=200,
            payload=json_data,
            repeat=True,
        )
        mock_aioresponse.post(
            url("/ivp/ss/dry_contact_settings"),
            status=200,
            payload=json_data,
            repeat=True,
        )

    if "ivp_ensemble_power" in files:
        try:
            json_data = await load_json_fixture(version, "ivp_ensemble_power")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/ivp/ensemble/power"), status=200, payload=json_data, repeat=True
        )

    if "ivp_ensemble_secctrl" in files:
        try:
            json_data = await load_json_fixture(version, "ivp_ensemble_secctrl")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/ivp/ensemble/secctrl"), status=200, payload=json_data, repeat=True
        )

    if "admin_lib_tariff" in files:
        try:
            json_data = await load_json_fixture(version, "admin_lib_tariff")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/admin/lib/tariff"), status=200, payload=json_data, repeat=True
        )
        mock_aioresponse.get(
            url_http("/admin/lib/tariff"), status=200, payload=json_data, repeat=True
        )
        mock_aioresponse.put(
            url("/admin/lib/tariff"), status=200, payload=json_data, repeat=True
        )
    else:
        mock_aioresponse.get(url("/admin/lib/tariff"), status=404, repeat=True)

    if "ivp_ss_gen_config" in files:
        try:
            json_data = await load_json_fixture(version, "ivp_ss_gen_config")
        except json.decoder.JSONDecodeError:
            json_data = {}
        mock_aioresponse.get(
            url("/ivp/ss/gen_config"), status=200, payload=json_data, repeat=True
        )
    else:
        mock_aioresponse.get(
            url("/ivp/ss/gen_config"), status=200, payload={}, repeat=True
        )

    if "home" in files:
        mock_aioresponse.get(
            url("/home"),
            status=200,
            payload=await load_json_fixture(version, "home"),
            repeat=True,
        )
    else:
        mock_aioresponse.get(url("/home"), status=404, repeat=True)

    if "ivp_pdm_device_data" in files:
        mock_aioresponse.get(
            url("/ivp/pdm/device_data"),
            status=200,
            payload=await load_json_fixture(version, "ivp_pdm_device_data"),
            repeat=True,
        )
    else:
        mock_aioresponse.get(url("/ivp/pdm/device_data"), status=404, repeat=True)

    return files


def updater_features(updaters: list[EnvoyUpdater]) -> dict[str, SupportedFeatures]:
    """Return the updater supported features flags"""
    return {type(updater).__name__: updater._supported_features for updater in updaters}



"""Model for the Envoy's production data."""

# Data Source: URL_PRODUCTION
from __future__ import annotations

from dataclasses import dataclass
from typing import Any


def find_dict_by_key(all_production: list[dict[str, Any]], key: str) -> dict[str, Any]:
    """Find a dict by key."""
    for production in all_production:
        if production.get("type") == key:
            return production
    raise ValueError(f"{key} is missing")


@dataclass(slots=True)
class EnvoySystemProduction:
    """Model for the Envoy's production data."""

    watt_hours_lifetime: int  #: Lifetime Energy produced
    watt_hours_last_7_days: (
        int  #: Energy produced in previous 7 days (not including today)
    )
    watt_hours_today: int  #: Energy produced since start of day
    watts_now: int  #: Current Power production

    @classmethod
    def from_v1_api(cls, data: dict[str, Any]) -> EnvoySystemProduction:
        """
        Initialize from the V1 API.

        :param data:  JSON reply from api/v1/production endpoint
        :return: Lifetime, last seven days, todays energy and current power for solar production
        """
        return cls(
            watt_hours_lifetime=data["wattHoursLifetime"],
            watt_hours_last_7_days=data["wattHoursSevenDays"],
            watt_hours_today=data["wattHoursToday"],
            watts_now=data["wattsNow"],
        )

    @classmethod
    def from_production(cls, data: dict[str, Any]) -> EnvoySystemProduction:
        """
        Initialize from the production API.

        :param data: JSON reply from /production endpoint
        :return: Lifetime, last seven days, todays energy and current power for solar production
        """
        all_production = data["production"]

        eim = find_dict_by_key(all_production, "eim")
        inverters = find_dict_by_key(all_production, "inverters")

        # This is backwards compatible with envoy_reader
        # envoy metered without configured CT has whLifetime and wNow in inverters
        # whLastSevenDays and whToday are incorrect for both so either can be used
        now_source = eim if eim["activeCount"] else inverters

        return cls(
            watt_hours_lifetime=round(now_source["whLifetime"]),
            watt_hours_last_7_days=round(
                eim.get("whLastSevenDays") or inverters.get("whLastSevenDays") or 0
            ),
            watt_hours_today=round(eim.get("whToday") or inverters.get("whToday") or 0),
            watts_now=round(now_source["wNow"]),
        )

    @classmethod
    def from_production_phase(
        cls, data: dict[str, Any], phase: int
    ) -> EnvoySystemProduction | None:
        """
        Initialize from the production API phase data.

        :param data: JSON reply from /production endpoint
        :param phase: Index (0-2) in [lines] segment for which to return data
        :return: Lifetime, last seven days, todays energy and current power for production phase
        """
        all_production = data["production"]
        eim = find_dict_by_key(all_production, "eim")

        # if {production[type=eim]{Lines[]} or phase is missing return None
        phases = eim.get("lines")
        if not phases or phase >= len(phases):
            return None

        phase_data = phases[phase]
        return cls(
            watt_hours_lifetime=round(phase_data.get("whLifetime") or 0),
            watt_hours_last_7_days=round(phase_data.get("whLastSevenDays") or 0),
            watt_hours_today=round(phase_data.get("whToday") or 0),
            watts_now=round(phase_data.get("wNow") or 0),
        )



import logging
from typing import Any

from ..const import ENSEMBLE_MIN_VERSION, URL_GEN_CONFIG, SupportedFeatures
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS
from ..models.envoy import EnvoyData
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyGeneratorUpdater(EnvoyUpdater):
    """Class to handle updates for Generator information."""

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this updater and return SupportedFeatures."""
        if self._envoy_version < ENSEMBLE_MIN_VERSION:
            _LOGGER.debug("Firmware too old for Ensemble support")
            return None

        # If there's no Enpower smart switch, we can skip the generator probe
        if SupportedFeatures.ENPOWER not in discovered_features:
            return None

        # Check for generator support
        try:
            result = await self._json_probe_request(URL_GEN_CONFIG)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Generator config endpoint not found: %s", e)
        else:
            if not result or "error" in result or "err" in result:
                # Newer firmware with no generator configured returns an empty dict
                _LOGGER.debug("No generator found")
                return None

            self._supported_features |= SupportedFeatures.GENERATOR

        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the generator data if supported."""
        generator_config_data: list[dict[str, Any]] = await self._json_request(
            URL_GEN_CONFIG
        )
        envoy_data.raw[URL_GEN_CONFIG] = generator_config_data



"""Model for the Envoy tariff data."""

from __future__ import annotations

from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class EnvoyStorageMode(StrEnum):
    BACKUP = "backup"
    SELF_CONSUMPTION = "self-consumption"
    SAVINGS = "economy"
    LEGACY_SAVINGS = "savings-mode"


@dataclass
class EnvoyTariff:
    """Model for the Envoy tariff data."""

    currency: dict[str, Any]
    logger: str | None
    date: str | None
    storage_settings: EnvoyStorageSettings | None
    single_rate: dict[str, Any]
    seasons: list[Any]
    seasons_sell: list[Any] | None

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> EnvoyTariff:
        """Initialize from the API."""
        return cls(
            currency=data["currency"],
            logger=data.get("logger"),
            date=data.get("date"),
            storage_settings=(
                EnvoyStorageSettings.from_api(data["storage_settings"])
                if data.get("storage_settings")
                else None
            ),
            single_rate=data["single_rate"],
            seasons=data["seasons"],
            seasons_sell=data.get("seasons_sell"),
        )

    def to_api(self) -> dict[str, Any]:
        """Convert to API format."""
        retval = {
            "currency": self.currency,
            "single_rate": self.single_rate,
            "seasons": self.seasons,
        }
        if self.logger:
            retval["logger"] = self.logger
        if self.date:
            retval["date"] = self.date
        if self.seasons_sell:
            retval["seasons_sell"] = self.seasons_sell
        if self.storage_settings:
            retval["storage_settings"] = self.storage_settings.to_api()

        return retval


@dataclass
class EnvoyStorageSettings:
    """Model for the Envoy storage settings."""

    mode: EnvoyStorageMode | None
    operation_mode_sub_type: str
    reserved_soc: float
    very_low_soc: int
    charge_from_grid: bool
    date: str | None
    opt_schedules: bool | None

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> EnvoyStorageSettings:
        """Initialize from the API."""
        return cls(
            # It appears a `mode` value of `economy` and `savings-mode` is interchangeable
            # However, the Enlighten app is using the `economy` value, so we will convert
            # `savings-mode` to `economy`
            # some fw return None in storage_settings["data"]
            mode=(
                EnvoyStorageMode.SAVINGS
                if data["mode"] == EnvoyStorageMode.LEGACY_SAVINGS.value
                else EnvoyStorageMode(data["mode"])
                if data.get("mode")
                else None
            ),
            operation_mode_sub_type=data["operation_mode_sub_type"],
            reserved_soc=data["reserved_soc"],
            very_low_soc=data["very_low_soc"],
            charge_from_grid=data["charge_from_grid"],
            # Some firmware versions don't return date
            date=data.get("date"),
            # opt_schedules was added in 8.2.42xx
            opt_schedules=data.get("opt_schedules"),
        )

    def to_api(self) -> dict[str, Any]:
        """Convert to API format."""
        retval = {
            "mode": self.mode.value if self.mode else None,
            "operation_mode_sub_type": self.operation_mode_sub_type,
            "reserved_soc": self.reserved_soc,
            "very_low_soc": self.very_low_soc,
            "charge_from_grid": self.charge_from_grid,
        }
        if self.date is not None:
            retval["date"] = self.date
        # opt_schedules was added in 8.2.42xx
        if self.opt_schedules is not None:
            retval["opt_schedules"] = self.opt_schedules

        return retval
